import pytest
import torch
import numpy as np
from ptlbp import DiffLBP, TOffsets
import ptlbp
from ptlbp.testing import generate_rect_image, generate_cross_image, generate_cross_and_small, \
    generate_square_and_small, generate_noise_image

from skimage.feature import local_binary_pattern
np.set_printoptions(linewidth=400)

# Debug flags, activate these to see the plots
debug_plots_forward = False
debug_plots_gradient = False
debug_test_generate_gradient = False
debug_plots_supress = False
debug_plots_scikit_image = False
debug_plots_gradient_ascent = False


@pytest.mark.parametrize("point_sample_str, supress_255,  supress_0, num_points, num_output_channels, histogram_size", [
    ("8r1", True, True, 8, 2**8, 2**8 - 2),
    ("8r1", False, True, 8, 2**8, 2**8 - 1),
    ("8r1", True, False, 8, 2**8, 2**8 - 1),
    ("8r1", False, False, 8, 2**8, 2**8),
    ("1r1", False, False, 1, 2**1, 2**1),
    ("3r1", False, False, 3, 2**3, 2**3),
])
def test_difflbp_properties(point_sample_str, supress_255,  supress_0, num_points, num_output_channels, histogram_size):
    lbp = DiffLBP(point_sample_str, supress_full_pattern=supress_255, supress_zero_pattern=supress_0)
    assert lbp.num_points == num_points
    assert lbp.num_output_channels == num_output_channels
    assert lbp.histogram_size == histogram_size


@pytest.mark.parametrize("image, offsets, hist_nonzero", [
    (generate_rect_image(100), "8r1", {0: 9400.0, 2: 1.0, 3: 1.0, 6: 1.0, 8: 1.0, 12: 1.0, 14: 48.0, 24: 1.0,
                                       32: 1.0, 48: 1.0, 56: 48.0, 96: 1.0, 128: 1.0, 129: 1.0, 131: 48.0, 192: 1.0,
                                       224: 48.0}),
    (generate_cross_image(100), "8r1", {0: 9400.0, 2: 2.0, 3: 2.0, 6: 2.0, 8: 2.0, 12: 2.0, 14: 44.0, 24: 2.0,
                                        32: 2.0, 48: 2.0, 56: 44.0, 62: 1.0, 96: 2.0, 128: 2.0, 129: 2.0, 131: 44.0,
                                        143: 1.0, 192: 2.0, 224: 44.0, 227: 1.0, 248: 1.0})
])
def test_DiffLBP_forward(image: torch.Tensor, offsets: TOffsets, hist_nonzero: dict):
    target_hist = torch.zeros(256)
    for k, v in hist_nonzero.items():
        target_hist[k] = v

    # Create an instance of DiffLBP
    diff_lbp = DiffLBP(offsets=offsets, diff_hardness=100., output_hardness=100.)

    # Perform forward pass
    output = diff_lbp(image)

    # Degug print and plot
    if debug_plots_forward:
        from matplotlib import pyplot as plt
        print(output.view(-1).detach().numpy().tolist())
        d = output.view(-1).detach().numpy().tolist()
        print({n: k for n, k in enumerate(d) if k > 0.0005})
        plt.plot(output.view(-1).detach().numpy().tolist())
        plt.show()

        # # Assert the output shape
        assert output.shape == (1, 256, 1, 1)

        # This is how the values for the test were generated
        # h = {}
        # for i in range(256):
        #     count = output[0, i, 0, 0].item()
        #     if count > 0.000000001:
        #         h[i] = count
        # print("H:", repr(h))

    assert torch.allclose(output.view(-1), target_hist.view(-1), atol=1e-7)


@pytest.mark.parametrize("image, offsets, erase_pattern, ignore_zero, small_grad, median_grad, large_grad", [
    (generate_cross_image(100), "8r1", 0, False, -0.8780800700187683, 0.0, 0.8780801296234131),
    (generate_cross_image(100), "8r1", 0, True, 0., 0., 0.),
    (generate_cross_image(100), "8r1", 1, True, -3.87195797e-07, 0., 3.3202590543e-07),
    (generate_cross_image(100), "8r1", 1, False, -3.87195797e-07, 0., 3.3202590543e-07),
    (generate_cross_image(100), "8r1", 2, True, -6.4000669e-05, 0., 6.4000647e-05),
    (generate_cross_image(100), "8r1", 79, True, 0., 0., 0.),
    (generate_cross_image(100), "8r1", 14, True, 0., 0., 0.),
])
def test_gradient(image: torch.Tensor, offsets: str, erase_pattern: int, ignore_zero: bool, small_grad: float,
                  median_grad: float, large_grad: float) -> None:
    diff_lbp = DiffLBP(offsets=offsets, diff_hardness=10., output_hardness=10., pool_pixels="sum")
    image.requires_grad_()
    target_hist = torch.zeros(256)
    output = diff_lbp(image).view(-1)
    target_hist = output.clone().detach()
    target_hist[erase_pattern] = 0
    if ignore_zero:
        loss = ((output[1:] - target_hist[1:])**2).sum()
    else:
        loss = ((output - target_hist)**2).sum()
    loss.backward()
    grad_img = image.grad
    out_small_grad = torch.quantile(grad_img, .02)
    out_median_grad = torch.quantile(grad_img, .5)
    out_large_grad = torch.quantile(grad_img, .98)
    if debug_plots_gradient:
        from matplotlib import pyplot as plt
        plt.imshow(image.grad[0, 0, :, :].detach().numpy())
        plt.colorbar()
        plt.title(f"Erased:{erase_pattern} Zero ignored:{ignore_zero}")
        plt.show()
    # This was used to generate the values
    # print(f"\n{erase_pattern} {ignore_zero}: {out_small_grad}, {out_median_grad}, {out_large_grad}")
    if ignore_zero:
        precision = 1e-5
    else:
        precision = 1e-1
    assert torch.allclose(out_small_grad, torch.tensor(small_grad, dtype=torch.float32), atol=precision)
    assert torch.allclose(out_median_grad, torch.tensor(median_grad, dtype=torch.float32), atol=precision*0.001)
    assert torch.allclose(out_large_grad, torch.tensor(large_grad, dtype=torch.float32), atol=precision)


def test_generate_gradient_image():
    image = generate_cross_image(100)
    diff_lbp = DiffLBP(offsets="8r1", diff_hardness=10, output_hardness=10)
    output = diff_lbp(image)
    # TODO(anguelos) find why removing the epsilon from the output gives a NaN image.
    # The epsilon can be assigned to a single cell of output and it still works
    weak_grad_image = diff_lbp.get_gradient_image(image, output + output * torch.rand_like(output) * .01)
    strong_grad_image = diff_lbp.get_gradient_image(image, torch.ones_like(output))
    if debug_plots_gradient_ascent:
        from matplotlib import pyplot as plt
        print("output: ", output.view(-1).detach().numpy().tolist())
        plt.imshow(strong_grad_image[0, 0, :, :].numpy())
        plt.colorbar()
        plt.title("Strong gradient")
        plt.show(block=False)
        plt.figure()
        plt.imshow(weak_grad_image[0, 0, :, :].numpy())
        plt.colorbar()
        plt.title("Weak gradient")
        plt.show()
    assert weak_grad_image.shape == strong_grad_image.shape == image.shape

    # Very rearelly the following fails randomly. This should be deterministic but
    # maybe numeric instability is not deterministic. Reducing 10 to 5 makes the test pass more often
    # assert (weak_grad_image ** 2).sum() * 10 < (strong_grad_image ** 2).sum()
    assert (weak_grad_image ** 2).sum() * 5 < (strong_grad_image ** 2).sum()
    grad_image = diff_lbp.get_gradient_image(image, torch.ones(256))
    assert grad_image.shape == image.shape

    # TODO(anguelos) test this more extencively
    grad_img = diff_lbp.get_triplet_gradient_image(image, torch.ones(1, 256, 1, 1), torch.ones(1, 256, 1, 1))
    assert grad_img.shape == image.shape
    grad_img = diff_lbp.get_triplet_gradient_image(image, torch.ones(256), torch.ones(256))
    assert grad_img.shape == image.shape


def test_supress():
    """Test the supression of zero and full patterns in the output histogram.
    """
    image = generate_cross_image(100)
    image[:, :, 5, 5] = 1
    image[:, :, 5, 94] = 1
    image[:, :, 94, 5] = 1
    image[:, :, 94, 94] = 1
    diff_lbp_nz_na = DiffLBP(offsets="8r1", diff_hardness=3, output_hardness=3,
                             supress_zero_pattern=True, supress_full_pattern=True)
    diff_lbp_nz_ya = DiffLBP(offsets="8r1", diff_hardness=3, output_hardness=3,
                             supress_zero_pattern=True, supress_full_pattern=False)
    diff_lbp_yz_na = DiffLBP(offsets="8r1", diff_hardness=3, output_hardness=3,
                             supress_zero_pattern=False, supress_full_pattern=True)
    diff_lbp_yz_ya = DiffLBP(offsets="8r1", diff_hardness=3, output_hardness=3,
                             supress_zero_pattern=False, supress_full_pattern=False)
    for variant in [diff_lbp_nz_na, diff_lbp_nz_ya, diff_lbp_yz_na, diff_lbp_yz_ya]:
        target = variant(torch.zeros_like(image))
        grad_image = variant.get_gradient_image(image, target)
        if debug_plots_supress:
            from matplotlib import pyplot as plt
            plt.imshow(grad_image[0, 0, :, :].detach().numpy() ** 2)
            plt.colorbar()
            plt.title(f"Supress zero: {variant.supress_zero_pattern}, Supress full: {variant.supress_full_pattern}")
            plt.show()
        assert grad_image.shape == image.shape


def get_jaccard(lbp_img1, lbp_img2, e=0.000000000000001):
    lab1_sz = int(lbp_img1.max() + 1)  # size of the img1 namespace
    lab2_sz = int(lbp_img2.max() + 1)  # size of the img2 namespace
    labs1, lab1_freqs = np.unique(lbp_img1, return_counts=True)
    labs2, lab2_freqs = np.unique(lbp_img2, return_counts=True)
    joint, joint_freqs = np.unique(lbp_img2.astype("uint32") * lab1_sz + lbp_img1.astype("uint32"), return_counts=True)
    lab1_hist = np.zeros(lab1_sz)
    lab1_hist[labs1.astype("uint32")] = lab1_freqs
    lab2_hist = np.zeros(lab2_sz)
    lab2_hist[labs2.astype("uint32")] = lab2_freqs
    joint_hist = np.zeros(lab1_sz * lab2_sz)
    joint_hist[joint] = joint_freqs
    sparse_union_table = np.zeros([lab1_sz, lab2_sz])
    sparse_intersection_table = np.zeros([lab1_sz, lab2_sz])
    for joint_lab in joint:
        lab1 = joint_lab % lab1_sz
        lab2 = joint_lab // lab1_sz
        sparse_union_table[lab1, lab2] = (lab1_hist[lab1] + lab2_hist[lab2]) - joint_hist[joint_lab]
        sparse_intersection_table[lab1, lab2] = joint_hist[joint_lab]

    # Getting the labels
    lab1 = np.arange(lab1_sz)        
    lab1 = lab1[lab1_hist > 0]
    lab2 = np.arange(lab2_sz)
    lab2 = lab2[lab2_hist > 0]
    intersection_table = sparse_intersection_table[lab1, :][:, lab2]
    union_table = sparse_union_table[lab1, :][:, lab2]

    #  Jaccard of two empty sets is defined as 0 therefor no e goes on the numerator
    return (intersection_table / (union_table + e), intersection_table, union_table,
            lab1, lab2, lab1_hist[lab1], lab2_hist[lab2])


parameters = [
              #  As the diff_harness goes down, the consistency between diff_lbp and sk_lbp goes down
              [100, 100, .97, 1, .78, 1],
              [10, 100, .97, 2, .78, 1],
              [3, 100, .66, 6, .78, 1],  # one patterns missed in a class where 3 occur (rotated 4 times)
              [1, 100, .66, 5, .78, 1],  # 6 becomes 5 because the 0 pattern is no longer created
              [.9, 100, .66, 5, .78, 1],
              [.5, 100, .66, 5, .78, 1],  # Eveything is mapped as the 255 pattern

              [100, .1, .97, 1, .78, 1],  # But we can see that the output_hardness has no effect
              ]


@pytest.mark.parametrize("diff_hardness, output_hardness, diff_min_precision, diff_imprecise_allowed,\
    sk_min_precision, sk_imprecise_allowed", parameters)
def test_scikit_image_agreement(diff_hardness, output_hardness, diff_min_precision,
                                diff_imprecise_allowed, sk_min_precision, sk_imprecise_allowed):
    # TODO (anguelos) merge this with the test in test_binary_to_onehot
    # Scikit image assigns different labels to each pattern but they should be internally coherent
    # A difference between our and the scikit variant is whether we use > cs . >= which allows for some confusion
    # between 0 and 255 patterns
    pt_img = generate_cross_and_small()
    np_img = pt_img.numpy()[0, 0, :, :]
    fulllbp = DiffLBP(offsets="8r1", diff_hardness=diff_hardness, output_hardness=output_hardness,
                      supress_zero_pattern=False, supress_full_pattern=False, pool_pixels=False)
    diff_lbp_img = torch.argmax(fulllbp(pt_img).detach(), dim=1)[0, :, :].numpy().astype("uint8")
    sk_lbp_img = local_binary_pattern(np_img.astype("uint8"), 8, 1, 'default')[1:-1, 1:-1]
    iou, i, u, diff_lab, sk_lab, diff_hist, sk_hist = get_jaccard(diff_lbp_img, sk_lbp_img)

    assert diff_hist.sum() == sk_hist.sum() == i.sum()  # sanity check about get_jaccard

    sk_precision = i.max(axis=0)/i.sum(axis=0)
    assert sk_hist.size == np.unique(sk_lbp_img).size  # sanity check about get_jaccard
    assert sk_precision.size == np.unique(sk_lbp_img).size  # sanity check about get_jaccard

    diff_precision = i.max(axis=1)/i.sum(axis=1)
    assert sk_hist.size == np.unique(sk_lbp_img).size   # sanity check about get_jaccard
    assert diff_precision.size == np.unique(diff_lbp_img).size  # sanity check about get_jaccard

    if debug_plots_scikit_image:
        print(f"Diff Hardnes: {diff_hardness}, Output Hardnes: {output_hardness}")
        print(f"diff_precision : {diff_precision}")
        print(f"sk_precision : {sk_precision}")

    assert diff_precision.min() > diff_min_precision
    assert (diff_precision < 1).sum() <= diff_imprecise_allowed
    assert sk_precision.min() > sk_min_precision
    assert (sk_precision < 1).sum() <= sk_imprecise_allowed

    if debug_plots_scikit_image:
        from matplotlib import pyplot as plt
        fig, axs = plt.subplots(2, 6)
        for n, pattern in enumerate([0, 1, 4, 16, 64, 255]):
            axs[0, n].imshow(diff_lbp_img == pattern)
            axs[1, n].imshow(sk_lbp_img == pattern)
        plt.show()
        plt.imshow(iou)
        plt.colorbar()
        plt.show()


def test_gradient_ascent():
    cross_img = generate_cross_and_small()
    square_img = generate_square_and_small()
    lbp = DiffLBP(offsets="8r1", diff_hardness=10, output_hardness=1,
                  supress_zero_pattern=True, supress_full_pattern=True)
    with torch.no_grad():
        cross_output = lbp(cross_img)
        square_output = lbp(square_img)
    square_grad = lbp.get_gradient_image(square_img, cross_output)
    cross_grad = lbp.get_gradient_image(cross_img, square_output)
    if debug_plots_gradient_ascent:
        from matplotlib import pyplot as plt
        grad_img = torch.cat([square_grad, cross_grad], dim=3)
        plt.imshow(grad_img[0, 0].numpy())
        plt.colorbar()
        plt.show()


@pytest.mark.parametrize("scale, translate, close_mean", [[1., 0., 1.],
                                                          [.7, 0., 1.], [.5, 0., .995],
                                                          [.3, 0., .971], [.1, 0., .75],
                                                          [1.5, 0., .996], [2., 0., 1.],
                                                          [1., .001, 1.], [1., -.5, .75],
                                                          ])
def test_ilumination_invarince(scale, translate, close_mean):
    pt_img = generate_cross_and_small() * .5
    fulllbp = DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                      supress_zero_pattern=False, supress_full_pattern=False, pool_pixels=False)
    gray_lbp_img = torch.argmax(fulllbp(pt_img).detach(), dim=1)[0, :, :].numpy().astype("uint8")
    shifted_lbp_img = torch.argmax(fulllbp(pt_img*scale+translate).detach(), dim=1)[0, :, :].numpy().astype("uint8")
    assert (gray_lbp_img == shifted_lbp_img).mean() >= close_mean


def test_gpu():
    lbp = ptlbp.DiffLBP("8r1")
    assert lbp.device == torch.device('cpu')
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    if torch.cuda.is_available():
        lbp.to('cuda')
        cuda_output = lbp(img.to('cuda'))
        assert torch.allclose(output, cuda_output.cpu(), atol=1e-6)
        assert str(lbp.device).split(':')[0] == 'cuda'
    else:
        with pytest.raises(RuntimeError):
            lbp.to('cuda')
            lbp(generate_noise_image(100, mode_balance=.5, seed=1337).to('cuda'))


def test_diff_pool():
    lbp = ptlbp.DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                        supress_zero_pattern=False, supress_full_pattern=False, pool_pixels="sum")
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    lbp = ptlbp.DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                        supress_zero_pattern=False, supress_full_pattern=False, pool_pixels="avg")
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    lbp = ptlbp.DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                        supress_zero_pattern=False, supress_full_pattern=False, pool_pixels="discrete")
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    lbp = ptlbp.DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                        supress_zero_pattern=False, supress_full_pattern=False,
                        pool_pixels=torch.nn.AdaptiveAvgPool2d(1))
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    lbp = ptlbp.DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                        supress_zero_pattern=False, supress_full_pattern=False, pool_pixels=None)
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 98, 98)

    with pytest.raises(ValueError):
        lbp = ptlbp.DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                            supress_zero_pattern=False, supress_full_pattern=False, pool_pixels="unknown")
        img = generate_noise_image(100, mode_balance=.5, seed=1337)
        output = lbp(img)
        assert output.shape == (1, 256, 1, 1)


def test_diff_comparissons():
    lbp = ptlbp.DiffLBP(offsets="8r1", comparisson="otsu")
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    lbp = ptlbp.DiffLBP(offsets="8r1", comparisson="simple")
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    lbp = ptlbp.DiffLBP(offsets="8r1", comparisson="quantile")
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    lbp = ptlbp.DiffLBP(offsets="8r1", comparisson=ptlbp.Comparisson())
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    assert output.shape == (1, 256, 1, 1)

    with pytest.raises(ValueError):
        lbp = ptlbp.DiffLBP(offsets="8r1", comparisson=None)
        img = generate_noise_image(100, mode_balance=.5, seed=1337)
        output = lbp(img)
        assert output.shape == (1, 256, 1, 1)


def test_compute_lbp_image():
    lbp = ptlbp.DiffLBP(offsets="8r1", diff_hardness=100, output_hardness=100,
                        supress_zero_pattern=False, supress_full_pattern=False, pool_pixels=False)
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    lbp_img, confidence = lbp.compute_lbp_image(img)
    patterns = np.unique(lbp_img.detach().numpy())
    assert 0 < len(patterns) <= 256
    assert 0 <= confidence.min() <= confidence.max() <= 1


def test_double_onehots():
    img = generate_cross_and_small()

    single = ptlbp.DiffLBP(offsets="8r1", double_onehots=False)
    double = ptlbp.DiffLBP(offsets="8r1", double_onehots=True)
    res_from_double = double(img).float()
    res_from_single = single(img).float()
    assert torch.allclose(res_from_double, res_from_single, atol=30)

    single = ptlbp.DiffLBP(offsets="8r1", double_onehots=False, pool_pixels="avg")
    double = ptlbp.DiffLBP(offsets="8r1", double_onehots=True, pool_pixels="avg")
    res_from_double = double(img).float()
    res_from_single = single(img).float()
    assert torch.allclose(res_from_double, res_from_single, atol=1e-3)


# TODO (anguelos) expand this to test for average pooling
@pytest.mark.parametrize("pool_pixels", ["discrete", "avg"])
def test_forward_sliced(pool_pixels):
    if pool_pixels in ["discrete"]:
        img = generate_cross_and_small()
        assert img.size(3) == img.size(2) == 100
        lbp = ptlbp.DiffLBP(offsets="8r1", pool_pixels=pool_pixels, output_hardness=100, diff_hardness=100)
        double_lbp = ptlbp.DiffLBP(offsets="8r1", double_onehots=True, pool_pixels=pool_pixels)
        output = lbp(img)
        with torch.no_grad():
            sliced_output = lbp.forward_sliced(img, 40)
            double_slice_output = double_lbp.forward_sliced(img, 40)
            pseudo_sliced_output = lbp.forward_sliced(img, 200)
        assert torch.allclose(output, sliced_output, atol=1e-2)
        assert torch.allclose(output, pseudo_sliced_output, atol=1e-2)
        assert torch.allclose(output, double_slice_output, atol=30.)
    else:
        lbp = ptlbp.DiffLBP(offsets="8r1", pool_pixels=pool_pixels, output_hardness=100, diff_hardness=100)
        print("Raisiong:", pool_pixels)
        with pytest.raises(ValueError):
            with torch.no_grad():
                _ = lbp.forward_sliced(generate_cross_and_small(), 40)


def test_block_normalise():
    img = generate_cross_and_small()
    lbp_supr_bn = ptlbp.DiffLBP(offsets="8r1", pool_pixels="discrete", output_hardness=100, diff_hardness=100,
                                supress_full_pattern=True, supress_zero_pattern=True, block_normalise=True)
    lbp_bn = ptlbp.DiffLBP(offsets="8r1", pool_pixels="discrete", output_hardness=100, diff_hardness=100,
                           supress_full_pattern=False, supress_zero_pattern=False, block_normalise=True)
    lbp = ptlbp.DiffLBP(offsets="8r1", pool_pixels="discrete", output_hardness=100, diff_hardness=100,
                        supress_full_pattern=False, supress_zero_pattern=False, block_normalise=False)

    output_supr_bn = lbp_supr_bn(img)
    output_bn = lbp_bn(img)
    output = lbp(img)
    assert output_bn.size() == output.size() and output.size() != output_supr_bn.size()
    assert torch.allclose(output_bn.sum(), output_supr_bn.sum(), atol=1e-7)
    assert torch.allclose(output_bn.sum(), torch.ones(1), atol=1e-7)
    valid_pixels = lbp.valid_pixels_size(img.size())
    valid_pixels = valid_pixels[2] * valid_pixels[3]
    assert .99 * valid_pixels <= output.sum().item() <= 1.01 * valid_pixels



import torch
import numpy as np
import re
from typing import TypeVar


TOffsets = TypeVar('TOffsets', int, list, np.ndarray, torch.Tensor, str)


class PointSampler(torch.nn.Module):
    @staticmethod
    def str2offsets(offsets: str) -> np.ndarray:
        """Converts a string to a numpy array of shape (N, 2) where N is the number of points.

        The string can be: "WNES", "3x3", "NrR", where N is a float or integer  representation of the
        number of points and r is the radius.
        """
        if offsets in ["WNES"]:
            return np.array([[-1, 0], [0, 1], [1, 0], [0, -1]])
        elif offsets == "3x3":
            return np.array([[-1, -1], [-1, 0], [-1, 1], [0, -1], [0, 1], [1, -1], [1, 0], [1, 1]])
        elif len(re.findall(r"^\d{1,2}r\d{1,2}(?:\.\d{1,2})?$", offsets)) == 1:
            num_points, r = [eval(k) for k in offsets.split("r")]
            angles = np.linspace(0, 2*np.pi, num_points, endpoint=False)
            # angles = angles + (np.pi/16) * (r-1)
            return np.array([[np.cos(a) * r, np.sin(a) * r] for a in angles])
        else:
            raise ValueError("Invalid string for offsets")

    @staticmethod
    def process_offsets(offsets: TOffsets) -> np.ndarray:
        """Converts a string, integer, list, to a numpy array of shape (N, 2) where N is the number of points.

        If the input is already a numpy array, it is returned as is.
        """
        if isinstance(offsets, torch.Tensor):
            return offsets.numpy()
        elif isinstance(offsets, np.ndarray):
            return offsets
        elif isinstance(offsets, list):
            return np.asarray(offsets)
        elif isinstance(offsets, int):
            angles = np.linspace(0, 2*np.pi, offsets, endpoint=False)
            return np.array([[np.cos(a), np.sin(a)] for a in angles])
        elif isinstance(offsets, str):
            return PointSampler.str2offsets(offsets)
        else:
            raise ValueError("Invalid type for offsets")

    def __init__(self, input_channels: int, offsets: TOffsets) -> None:
        super().__init__()
        offsets = PointSampler.process_offsets(offsets)
        X = offsets[:, 0]
        Y = offsets[:, 1]
        num_points = offsets.shape[0]
        left_edge = int(np.floor(X).min())
        right_edge = int(np.ceil(X).max())
        top_edge = int(np.floor(Y).min())
        bottom_edge = int(np.ceil(Y).max())
        kernel_size = (1 + right_edge - left_edge, 1 + bottom_edge - top_edge)
        W = np.zeros((num_points*input_channels, input_channels, kernel_size[0], kernel_size[1]), dtype=np.float32)

        for n in range(num_points):
            left = int(np.floor(offsets[n, 0]))
            left_coef = 1 - (offsets[n, 0]-left)
            if left == offsets[n, 0]:
                right = left
                left_coef = .5
                right_coef = .5
            else:
                right = left + 1
                right_coef = 1 - left_coef

            # centering convolutions
            left = left - left_edge
            right = right - left_edge

            top = int(np.floor(offsets[n, 1]))
            top_coef = 1 - (offsets[n, 1]-top)
            if top == offsets[n, 1]:
                bottom = top
                bottom_coef = .5
                top_coef = .5
            else:
                bottom = top + 1
                bottom_coef = 1 - top_coef

            # centering convolutions
            top = top - top_edge
            bottom = bottom - top_edge

            repr([n, 0, top, left])
            W[n, :, left, top] += top_coef * left_coef
            W[n, :, right, top] += top_coef * right_coef
            W[n, :, left, bottom] += bottom_coef * left_coef
            W[n, :, right, bottom] += bottom_coef * right_coef

        self.weights = torch.nn.parameter.Parameter(torch.tensor(W))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.conv2d(x, self.weights)

    def render_pattern(self, pattern: int, res_width, res_height) -> torch.Tensor:
        with torch.no_grad():
            positive_bin_vector = torch.tensor([pattern // 2**k % 2 for k in range(self.num_points)]).bool()
            negative_bin_vector = ~positive_bin_vector
            res = torch.zeros((1, 1, res_height, res_width), dtype=torch.float32)
            positve_filters = self.weights[positive_bin_vector].sum(dim=0, keepdim=True)
            negative_filters = self.weights[negative_bin_vector].sum(dim=0, keepdim=True)
            assert res.size(2) >= positve_filters.size(2) and res.size(3) >= positve_filters.size(3)
            assert res.size(2) % 2 == 1 and res.size(3) % 2 == 1
            assert positve_filters.size(2) % 2 == 1 and positve_filters.size(3) % 2 == 1
            pad_x = (res.size(2) - positve_filters.size(2)) // 2
            pad_y = (res.size(3) - positve_filters.size(3)) // 2
            res[:, :, pad_x:-pad_x, pad_y:-pad_y] = positve_filters - negative_filters
            return res[0, 0, :, :].detach()

    def valid_pixels_size(self, input_size: tuple) -> tuple:
        border_crop_x = (self.weights.size(2) - 1)
        border_crop_y = (self.weights.size(3) - 1)
        return (input_size[0], self.weights.size(0), input_size[2] - border_crop_x, input_size[3] - border_crop_y)

    @property
    def num_points(self) -> int:
        return self.weights.size(0)//self.weights.size(1)

    @property
    def num_input_channels(self) -> int:
        return self.weights.size(1)


class GaussianPointSampler(PointSampler):
    @staticmethod
    def generate_gaussian_filters(X, Y, point_sigma, point_sum, input_channels=1):
        num_points = X.shape[0]
        if point_sigma < 0:
            point_sigma = X.size(0)
        left_edge = int(np.floor(X - point_sigma).min())
        right_edge = int(np.ceil(X + point_sigma).max())
        top_edge = int(np.floor(Y - point_sigma).min() - point_sigma)
        bottom_edge = int(np.ceil(Y + point_sigma).max())
        kernel_size = (1 + right_edge - left_edge, 1 + bottom_edge - top_edge)
        kernel_orgin = -left_edge, -top_edge
        print((X + point_sigma))
        print(f"X:{X}, Left:{left_edge} Right:{right_edge}, SZ:{kernel_size[0]}\nY:{Y}, Top:{top_edge} Bottom:{bottom_edge}, {kernel_size[1]}")
        W = np.zeros((num_points*input_channels, input_channels, kernel_size[0], kernel_size[1]), dtype=np.float32)
        return W

    def __init__(self, input_channels: int, offsets: TOffsets, point_sigma: float = -1., point_sum: float = 1.) -> None:
        super().__init__(input_channels=input_channels, offsets=offsets)
        offsets = PointSampler.process_offsets(offsets)
        X = offsets[:, 0]
        Y = offsets[:, 1]
        W = GaussianPointSampler.generate_gaussian_filters(X, Y, point_sigma=point_sigma, point_sum=point_sum, input_channels=1)
        self.weights = torch.nn.parameter.Parameter(torch.tensor(W))



import torch
from typing import Tuple, Union, Literal, Generator
from .sampling import PointSampler, TOffsets
from .comparisson import Comparisson, ComparissonQuantile, ComparissonOtsu
from .encoding import BinaryToOnehotChannelEmbeddings
import sys


class DiffLBP(torch.nn.Module):
    class DiscretePool(torch.nn.Module):
        def __init__(self, channels: int = 256) -> None:
            super().__init__()
            self.channels = channels

        def forward(self, x) -> torch.Tensor:
            x = torch.argmax(x, dim=1).float()[:, None, :, :]
            histograms = []
            for batch_item in x:
                histograms.append(torch.histc(batch_item, bins=self.channels, min=0, max=self.channels-1))
            return torch.stack(histograms, dim=0)[:, :, None, None]

    def __init__(self, offsets: TOffsets,
                 comparisson: Union[Literal["quantile", "otsu", "simple"], torch.nn.Module] = "quantile",
                 pool_pixels: Union[torch.nn.Module, Literal['sum', False, None, 'avg', 'discrete']] = 'sum',
                 diff_hardness: float = 3.0, learn_input_hardness: bool = False,
                 output_hardness: float = 3.0, learn_output_hardness: bool = False,
                 supress_full_pattern: bool = False, supress_zero_pattern: bool = False,
                 block_normalise: bool = False, double_onehots: bool = False) -> None:
        super().__init__()
        self.point_sampler = PointSampler(input_channels=1, offsets=offsets)

        if comparisson == "otsu":
            self.comparisson = ComparissonOtsu()
        elif comparisson == "simple":
            self.comparisson = Comparisson()
        elif comparisson == "quantile":
            self.comparisson = ComparissonQuantile()
        elif isinstance(comparisson, torch.nn.Module):
            self.comparisson = comparisson
        else:
            raise ValueError("Invalid comparisson")

        self.binary_to_onehot = BinaryToOnehotChannelEmbeddings(self.point_sampler.num_points)

        if pool_pixels == 'sum':
            self.histogram_pool = lambda x: x.sum(dim=2, keepdim=True).sum(dim=3, keepdim=True)
            self.pool_mode = 'sum'

        elif pool_pixels == 'avg':
            self.histogram_pool = torch.nn.AdaptiveAvgPool2d(1)
            self.pool_mode = 'avg'

        elif pool_pixels == 'discrete':
            self.histogram_pool = DiffLBP.DiscretePool(self.binary_to_onehot.num_output_channels)
            self.pool_mode = 'discrete'

        elif pool_pixels is False or pool_pixels is None:
            self.histogram_pool = lambda x: x
            self.pool_mode = None

        elif isinstance(pool_pixels, torch.nn.Module):
            self.histogram_pool = pool_pixels
            self.pool_mode = 'custom'

        else:
            raise ValueError("Invalid pool_pixels")

        self.diff_hardness = torch.nn.parameter.Parameter(torch.tensor(diff_hardness, dtype=torch.float32))
        self.diff_hardness.requires_grad = learn_input_hardness
        self.output_hardness = torch.nn.parameter.Parameter(torch.tensor(output_hardness, dtype=torch.float32))
        self.output_hardness.requires_grad = learn_output_hardness
        self.supress_full_pattern = supress_full_pattern
        self.supress_zero_pattern = supress_zero_pattern
        self.block_normalise = block_normalise
        self.double_onehots = double_onehots

    def compute_lbp_image(self, x: torch.Tensor, slice_width: int = 500,
                          return_confidences=True) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        with torch.no_grad():
            pad = self.point_sampler.weights.size(2) // 2
            pattern_image_slices = []
            confidence_image_slices = []
            for slice, _, _ in self.__forward_pixels_sliced(x, slice_width):
                confidences, patterns = slice.max(dim=1, keepdim=True)
                pattern_image_slices.append(patterns)
                confidence_image_slices.append(confidences)
            pattern_x = torch.cat(pattern_image_slices, dim=3)
            conf_x = torch.cat(confidence_image_slices, dim=3)
            res_pattern_img = torch.zeros_like(x)
            res_confidence_img = torch.zeros_like(x)
            res_pattern_img[:, :, pad:-pad, pad:-pad] = pattern_x
            res_confidence_img[:, :, pad:-pad, pad:-pad] = conf_x
        if return_confidences:
            return res_pattern_img, res_confidence_img
        else:
            return res_pattern_img

    def __forward_pixels(self, x: torch.Tensor) -> torch.Tensor:
        """Computes the forward pass of the model on pixels.

            This is the most memory inefficient part of the the layer.
        """
        x = self.comparisson(x, self.point_sampler(x))
        x = torch.nn.functional.tanh(x * self.diff_hardness) * .5 + .5
        x = self.binary_to_onehot(x)
        if self.double_onehots:
            x = x.double()
        x = torch.nn.functional.softmax(x * self.output_hardness, dim=1)
        return x

    def __forward_pixels_sliced(self, x: torch.Tensor, slice_width: int = 100)\
            -> Generator[Tuple[torch.Tensor, int, int], None, None]:
        """Computes the forward pass of the model on pixels."""
        assert not torch.is_grad_enabled()
        x = self.comparisson(x, self.point_sampler(x))
        x = torch.nn.functional.tanh(x * self.diff_hardness) * .5 + .5
        for i in range(0, x.size(3), slice_width):
            patch_x = self.binary_to_onehot(x[:, :, :, i:i+slice_width])
            pixel_count = patch_x.size(2) * patch_x.size(3)
            if self.double_onehots:
                print("Warning: DiffLbp.forward_pixels_sliced slicing an image and employing\
                    double precision might be wrong.", sys.stderr)
                patch_x = patch_x.double()
            patch_x = torch.nn.functional.softmax(patch_x * self.output_hardness, dim=1)
            yield patch_x, pixel_count, x.size(2) * x.size(3)

    def __forward_histograms(self, x: torch.Tensor) -> torch.Tensor:
        """Postprocesses the histograms after pooling."""
        if self.supress_full_pattern:
            x = x[:, :-1]
        if self.supress_zero_pattern:
            x = x[:, 1:]
        if self.block_normalise:
            x = x / (x.sum(dim=1, keepdim=True) + 1e-10)
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """This is the forward pass of the model.

        This method:
            computes patterns for every pixel,
            pools pixels into histograms,
            and histograms normalises resulting histograms.
        """
        x = self.__forward_pixels(x)
        x = self.histogram_pool(x)
        x = self.__forward_histograms(x)
        return x

    def forward_sliced(self, x: torch.Tensor, patch_width=1000) -> torch.Tensor:
        # TODO(anguelos) add support for vertical spliting of the image
        # TODO(anguelos) cache the forward pass to enable backward as well
        """Computes the forward pass on images that can't fit memory.

            Attention!!! this does not cache for the backward pass, so it is not suitable for training.
        """
        assert not torch.is_grad_enabled()

        # TODO(anguelos) add support for and None, sum pooling.
        if self.pool_mode not in ["discrete"]:
            raise ValueError("forward_sliced only supports 'discrete' pooling.")
        if patch_width >= x.size(3):
            return self.forward(x)
        x_accumulation = []
        for patch_x, pixel_count, total_forward_pixels in self.__forward_pixels_sliced(x, patch_width):
            patch_x = self.histogram_pool(patch_x)
            x_accumulation.append(patch_x)
        x = torch.cat(x_accumulation, dim=3).sum(dim=3, keepdim=True)
        x = self.__forward_histograms(x)
        return x

    def get_gradient_image(self, image: torch.Tensor, desired_output: torch.Tensor, e=1e-10) -> torch.Tensor:
        """
        Calculates the gradient of the loss function with respect to the input image.

        Args:
            image (torch.Tensor): The input image.
            desired_output (torch.Tensor): The desired output of the model either with dimensions BxCx1x1 or as with a
                                           single dimesion C equal to the histgram size.

        Returns:
            torch.Tensor: The gradient of the loss function with respect to the input image.
        """
        image = image.clone()
        des_output = desired_output.detach()
        tmp_image = image.clone()
        tmp_image.requires_grad_()
        output = self(tmp_image)
        if len(desired_output.size()) == 1:
            des_output = des_output[None, :, None, None]
        loss = (((output - desired_output)**2)**.5).sum()
        # TODO(anguelos) handle when the loss is 0, we get NaN in the image
        loss.backward()
        result = tmp_image.grad.detach()
        return result

    def get_triplet_gradient_image(self, ancor_image: torch.Tensor,
                                   positive_output: torch.Tensor, negative_output) -> torch.Tensor:
        """
        Calculates the gradient of the loss function with respect to the input image.

        Args:
            image (torch.Tensor): The input image.
            desired_output (torch.Tensor): The desired output of the model either with dimensions BxCx1x1 or as with a
                                           single dimesion C equal to the histgram size.

        Returns:
            torch.Tensor: The gradient of the loss function with respect to the input image.
        """
        ancor_image = ancor_image.clone()
        positive_output = positive_output.detach()
        negative_output = negative_output.detach()
        ancor_image.requires_grad_()
        output = self(ancor_image)
        if len(positive_output.size()) == 1:
            positive_output = positive_output[None, :, None, None]
        if len(negative_output.size()) == 1:
            negative_output = negative_output[None, :, None, None]

        positive_output = positive_output[:, 1:-1, :, :]
        negative_output = negative_output[:, 1:-1, :, :]
        output = output[:, 1:-1, :, :]

        positive_diff = (((output - positive_output)**2)).sum()
        negative_diff = (((output - negative_output)**2)).sum()
        loss = (((negative_diff-positive_diff)))
        print(f"Positive: {positive_diff.item():.20f} Negative: {negative_diff.item():.20f} Loss: {loss.item():.20f}")
        loss.backward()
        result = ancor_image.grad.detach()
        return result

    def valid_pixels_size(self, input_size: tuple) -> tuple:
        border_crop_x = (self.point_sampler.weights.size(2) - 1)
        border_crop_y = (self.point_sampler.weights.size(3) - 1)
        return (input_size[0], 2 ** self.point_sampler.weights.size(0),
                input_size[2] - border_crop_x, input_size[3] - border_crop_y)

    @property
    def num_points(self):
        return self.point_sampler.num_points

    @property
    def num_output_channels(self):
        return 2**self.num_points * self.point_sampler.num_input_channels

    @property
    def histogram_size(self):
        return self.num_output_channels - int(self.supress_full_pattern) - int(self.supress_zero_pattern)

    @property
    def device(self):
        return next(self.parameters()).device



import torch
from .histograms import hard_otsu_threshold


class Comparisson(torch.nn.Module):
    def __init__(self, margin: float = .25, learn_margin: bool = False) -> None:
        super().__init__()
        self.margin = torch.nn.parameter.Parameter(torch.tensor(margin, dtype=torch.float32))
        self.margin.requires_grad = learn_margin

    def forward(self, x, offset: torch.Tensor) -> torch.Tensor:
        crop_x = (x.size(2) - offset.size(2)) // 2
        crop_y = (x.size(3) - offset.size(3)) // 2
        return offset - x[:, :, crop_x:-crop_x, crop_y:-crop_y] - self.margin


class ComparissonOtsu(torch.nn.Module):
    def __init__(self,  bins: int = 255, assumed_max: float = 1.0) -> None:
        super().__init__()
        self.bins = bins
        self.assumed_max = assumed_max

    def forward(self, x, offset: torch.Tensor) -> torch.Tensor:
        crop_x = (x.size(2) - offset.size(2)) // 2
        crop_y = (x.size(3) - offset.size(3)) // 2
        diff = offset - x[:, :, crop_x:-crop_x, crop_y:-crop_y]
        # adding 1% noise to avoid quantile problems
        noisy_diff = diff + torch.randn_like(diff) * .01 * (diff.max() - diff.min())
        hist = torch.histc(noisy_diff, bins=self.bins, min=0, max=self.assumed_max)
        quantile_bin_up = (hist.cumsum(0)/hist.sum() < .99).sum().item()
        quantile_bin_down = hist.size(0) - (hist.cumsum(0)/hist.sum() > .01).sum().item()
        # enforce at lest 10 bins to be passed to the histogram
        if quantile_bin_down + 10 >= quantile_bin_up:
            quantile_bin_down -= 10
            quantile_bin_up += 10
            quantile_bin_down = max(1, quantile_bin_down)
            quantile_bin_up = min(self.bins - 1, quantile_bin_up)
        otsu_bin_idx = hard_otsu_threshold(hist[quantile_bin_down:quantile_bin_up]).item() + quantile_bin_down
        #print(otsu_bin_idx)
        #print(self.bins)
        margin = (self.assumed_max / self.bins) * otsu_bin_idx
        #margin = (otsu_bin * self.assumed_max) / self.bins
        #print(margin)
        return diff - margin


class ComparissonQuantile(torch.nn.Module):
    def __init__(self,  bins: int = 255, assumed_max: float = 1.0) -> None:
        super().__init__()
        self.bins = bins
        self.assumed_max = assumed_max

    def forward(self, x, offset: torch.Tensor) -> torch.Tensor:
        crop_x = (x.size(2) - offset.size(2)) // 2
        crop_y = (x.size(3) - offset.size(3)) // 2
        diff = offset - x[:, :, crop_x:-crop_x, crop_y:-crop_y]
        # adding 1% noise to avoid quantile problems
        noisy_diff = diff + torch.randn_like(diff) * .01 * (diff.max() - diff.min())
        hist = torch.histc(noisy_diff, bins=self.bins, min=0, max=self.assumed_max)
        quantile_bin_up = (hist.cumsum(0)/hist.sum() < .99).sum().item()
        quantile_bin_down = hist.size(0) - (hist.cumsum(0)/hist.sum() > .01).sum().item()
        margin = ((quantile_bin_down + quantile_bin_up) / 2) * self.assumed_max / self.bins
        return diff - margin



import numpy as np
import torch
import pytest
from ptlbp import PointSampler


@pytest.mark.parametrize("offsets, num_points", [([[0, 1], [1, 0]], 2),
                                                 (torch.Tensor([[0, 1], [1, 0]]), 2),
                                                 (np.array([[0, 1], [1, 0]]), 2),
                                                 ("8.5r1", -1),
                                                 ("8r1.5", 8),
                                                 ("3x3", 8),
                                                 (8, 8),
                                                 (5, 5),
                                                 ("WNES", 4),
                                                 (torch.rand(7, 2), 7),
                                                 (np.zeros([3, 2]), 3)])
def test_PointSampler_offsets(offsets, num_points):
    if num_points == -1:
        with pytest.raises(ValueError):
            sampler = PointSampler(1, offsets)
    else:
        sampler = PointSampler(1, offsets)
        assert sampler.weights.shape[0] == num_points
        assert sampler.weights.shape[1] == 1
        assert len(sampler.weights.size()) == 4


@pytest.mark.parametrize("offsets, kernel_size", [
    (torch.tensor([[0.5, 0.5], [1.5, 1.5]]), (3, 3)),
    (torch.tensor([[0.2, 0.2], [0.8, 0.8]]), (2, 2)),
    (torch.tensor([[0.1, 0.1], [0.3, 0.3], [0.5, 0.5], [0.7, 0.7], [-2, -1.5]]), (4, 4))
])
def test_PointSampler_sizes(offsets, kernel_size):
    input_channels = 3
    num_points = offsets.shape[0]
    expected_weight_shape = (num_points * input_channels, input_channels, kernel_size[0], kernel_size[1])
    sampler = PointSampler(input_channels, offsets)
    weight = sampler.weights.detach().numpy()
    assert weight.shape == expected_weight_shape


@pytest.mark.parametrize("input_channels, offsets, impulse_sum", [
    (1, torch.tensor([[0.0, 0.0]]), 1.),
    (1, torch.tensor([[0.5, 0.5], [1.5, 1.5]]), 2.),
    (3, torch.tensor([[0.2, 0.2], [0.8, 0.8]]), 6.),
    (3, torch.tensor([[0.1, 0.1], [0.3, 0.3], [0.5, 0.5], [0.7, 0.7]]), 12.)
])
def test_PointSampler_impulse_sum(input_channels, offsets, impulse_sum):
    sampler = PointSampler(input_channels, offsets)
    image = torch.zeros((1, input_channels, 10, 10))
    image[0, :, 3, 3] = 1.
    output = sampler(image).detach().numpy()
    output_sum = output.sum().item()
    assert np.allclose(output_sum, impulse_sum)


def test_str2offsets():
    simple_r1 = torch.tensor(PointSampler.str2offsets("8r1"), dtype=torch.float32)
    partial_target = torch.tensor([[1, 0], [0, 1], [-1, 0], [0, -1]], dtype=torch.float32)
    assert simple_r1.size(0) == 8 and simple_r1.size(1) == 2
    assert torch.allclose(simple_r1[[0, 2, 4, 6], :], partial_target)  # E,, N,, W,, S,

    v1 = torch.tensor(PointSampler.str2offsets("3x3"), dtype=torch.float32)
    v1_target = [[-1, -1], [-1, 0], [-1, 1], [0, -1], [0, 1], [1, -1], [1, 0], [1, 1]]
    assert torch.allclose(v1, torch.tensor(v1_target, dtype=torch.float32))

    large_r1 = torch.tensor(PointSampler.str2offsets("16r1"), dtype=torch.float32)
    partial_target = torch.tensor([[1, 0], [0, 1], [-1, 0], [0, -1]], dtype=torch.float32)
    assert large_r1.size(0) == 16 and large_r1.size(1) == 2
    assert torch.allclose(large_r1[[0, 4, 8, 12], :], partial_target)  # E,, N,, W,, S,

    simple_r2 = torch.tensor(PointSampler.str2offsets("8r2"), dtype=torch.float32)
    partial_target = torch.tensor([[1, 0], [0, 1], [-1, 0], [0, -1]], dtype=torch.float32) * 2
    assert simple_r2.size(0) == 8 and simple_r2.size(1) == 2
    assert torch.allclose(simple_r2[[0, 2, 4, 6], :], partial_target)  # E,, N,, W,, S,



#!/usr/bin/env python

from PIL import Image
import os
import os.path
import errno
import torch
from io import BytesIO
from torchvision import transforms
import glob
from pathlib import Path
import numpy as np
from collections import defaultdict
import sys
from typing import Tuple, Dict, List, Set
from abc import ABC, abstractmethod
from numpy.random import MT19937, RandomState, SeedSequence
import zipfile
import rarfile
import py7zr
import torchvision
from subprocess import getoutput as shell_stdout
import os


transform_384_train = transforms.Compose([
    transforms.RandomCrop((1024, 384), padding=1, pad_if_needed=True),
    transforms.Resize((512, 192)),
    transforms.Grayscale(),
    transforms.ToTensor(),
])

transform_test = transforms.Compose([
    transforms.Grayscale(),
    transforms.ToTensor(),
])


def read_rar_file(rarstream):
    names = rarstream.namelist()
    sample_list = []
    for name in names:
        img = Image.open(BytesIO(rarstream.read(rarstream.getinfo(name)))).copy()
        [writer_id, sample_id] = [int(s) for s in name[:name.find(".")].split("_")]
        language_id = (sample_id-1) / 2  # icdar 2013: english 1,2 and greek 3,4
        sample_list.append((img, (writer_id, language_id, sample_id), name))
    return sample_list


class TextureDs(ABC):
    @abstractmethod
    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int, str]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def get_labels(self) -> np.array:
        """Returns the labels (int classes) of each item in the dataset as a numpy array."""
        pass

    @abstractmethod
    def get_ids(self) -> np.array:
        """Returns the string identifiers of each item in the dataset as a numpy array."""
        pass

    def get_ids_to_indexes(self) -> Dict[str, int]:
        return {id: n for n, id in enumerate(self.get_ids())}

    def get_indexes_by_labels(self) -> Dict[int, List[int]]:
        res = defaultdict(lambda: [])
        for n, label in enumerate(self.get_labels()):
            res[label].append(n)
        return res

    def get_unique_labels(self) -> Set[str]:
        return set(self.get_indexes_by_labels().keys())


class TextureSubsetDs(TextureDs):
    def __init__(self, ds: TextureDs, indexes: List[int]):
        self.ds = ds
        self.item_indexes = np.array(indexes)

    def get_ids(self) -> np.array:
        return self.ds.get_ids()[self.item_indexes]

    def get_labels(self) -> np.array:
        return self.ds.get_labels()[self.item_indexes]

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int, str]:
        return self.ds[self.item_indexes[index]]

    def __len__(self) -> int:
        return self.item_indexes.size


class GenericTextureDs(TextureDs):
    def __init__(self, items: List[Tuple[torch.Tensor, int, str]], return_dims: int = 3) -> None:
        super().__init__()
        self.labels = np.array([label for _, label, _ in items])
        self.ids = np.array([id for _, _, id in items])
        if return_dims == 4:
            if items[0][0].dim() == 4:  # assuming batch of images
                self.inputs = [input for input, _, _ in items]
            elif items[0][0].dim() == 3:  # assuming image sample
                self.inputs = [input[None, :, None, None] for input, _, _ in items]
            elif items[0][0].dim() == 2:  # assuming batch of embeddings
                self.inputs = [input[:, :, None, None] for input, _, _ in items]
            elif items[0][0].dim() == 1:
                self.inputs = [input[None, :, None, None] for input, _, _ in items]
            else:
                raise ValueError(f"Invalid input dimension {items[0][0].dim()}")
        elif return_dims == 3:  # assuming image sample
            if items[0][0].dim() == 4:  # assuming batch of images
                if not all(input.size(0) == 1 for input, _, _ in items):
                    raise ValueError("Invalid batch size")
                self.inputs = [input[0] for input, _, _ in items]
            elif items[0][0].dim() == 3:  # assuming image sample
                self.inputs = [input for input, _, _ in items]
            elif items[0][0].dim() == 2:  # assuming batch of embeddings
                if not all(input.size(0) == 1 for input, _, _ in items):
                    raise ValueError("Invalid batch size")
                self.inputs = [input[0, :, None, None] for input, _, _ in items]
            elif items[0][0].dim() == 1:
                self.inputs = [input[None, :, None, None] for input, _, _ in items]
            else:
                raise ValueError(f"Invalid input dimension {items[0][0].dim()}")
        elif return_dims == 2:  # assuming embeddings
            if items[0][0].dim() == 4:
                if not all(input.size(1) == 1 and input.size(2) == 1 for input, _, _ in items):
                    raise ValueError("Invalid batch size")
                self.inputs = [input[:, :, 0, 0] for input, _, _ in items]
            elif items[0][0].dim() == 3:
                if not all(input.size(1) == 1 and input.size(2) == 1 for input, _, _ in items):
                    raise ValueError("Invalid batch size")
                self.inputs = [input[None, :, 0, 0] for input, _, _ in items]
            elif items[0][0].dim() == 2:
                self.inputs = [input for input, _, _ in items]
            elif items[0][0].dim() == 1:
                self.inputs = [input[None, :] for input, _, _ in items]
            else:
                raise ValueError(f"Invalid input dimension {items[0][0].dim()}")
        elif return_dims == 1:
            if items[0][0].dim() == 1:
                self.inputs = [input for input, _, _ in items]
            elif items[0][0].dim() == 2:
                self.inputs = [input[0, :] for input, _, _ in items]
            elif items[0][0].dim() == 3:
                if items[0][0].size(0) == 1:
                    print("Warning: single channel image should not be treated as an embedding", file=sys.stderr)
                if not all(input.size(1) == 1 and input.size(2) == 1 for input, _, _ in items):
                    raise ValueError("Invalid batch size")
                self.inputs = [input[:, 0, 0] for input, _, _ in items]
            elif items[0][0].dim() == 4:
                if items[0][0].size(1) == 1:
                    print("Warning: single channel image should not be treated as an embedding", file=sys.stderr)
                if not all(input.size(2) == 1 and input.size(3) == 1 and input.size(0) == 1 for input, _, _ in items):
                    raise ValueError("Invalid batch size")
                self.inputs = [input[0, :, 0, 0] for input, _, _ in items]
            else:
                raise ValueError(f"Invalid input dimension {items[0][0].dim()}")
        self.return_dims = return_dims

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int, str]:
        return self.inputs[index], self.labels[index], self.ids[index]

    def __len__(self) -> int:
        return len(self.inputs)

    def get_labels(self) -> np.array:
        return self.labels

    def get_ids(self) -> np.array:
        return self.ids


def split_by_labels(ds: TextureDs, label_counts: List[int], seed=-1) -> List[TextureSubsetDs]:
    unique_labels = np.array(list(ds.get_unique_labels()))
    if seed >= 0:
        rs = RandomState(MT19937(SeedSequence(seed)))
        unique_labels = rs.permutation(unique_labels)
    else:
        unique_labels = np.random.permutation(unique_labels)
    indexes_by_labels = ds.get_indexes_by_labels()
    datasets = []
    for label_count in label_counts:
        if label_count == -1:
            return_unique_labels = unique_labels
        else:
            return_unique_labels = unique_labels[:label_count]
            unique_labels = unique_labels[label_count:]
        result_indexes = []
        for label in return_unique_labels:
            result_indexes.extend(indexes_by_labels[label])
        datasets.append(TextureSubsetDs(ds, result_indexes))
    return datasets


class WI2017(TextureDs):
    def __init__(self, ds_root, img_folder="ScriptNet-HistoricalWI-2017-color", transform=None) -> None:
        super().__init__()
        self.transform = transform
        if transform is None:
            transform = transform_test
        self.ids = []
        self.labels = []
        for path in glob.glob(f"{ds_root}/{img_folder}/*.jpg"):
            name = path.split("/")[-1]
            self.ids.append(name)
            self.labels.append(int(name.split("-")[0]))
        print(f"WI2017 Loading {len(self.ids)} samples.", file=sys.stderr)
        self.labels = np.array(self.labels)
        self.ids = np.array(self.ids)
        self.root = ds_root
        self.image_path_pattern = f"{ds_root}/{img_folder}/{{}}"
        self.transform = transform

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int, str]:
        img = Image.open(self.image_path_pattern.format(self.ids[index]))
        img = self.transform(img)
        return img, self.labels[index], self.ids[index]

    def __len__(self) -> int:
        return len(self.ids)

    def get_ids(self) -> np.array:
        return self.ids.copy()

    def get_labels(self) -> np.array:
        return self.labels.copy()


class WI2019(TextureDs):
    def __init__(self, ds_root, train_saples_folder="wi_comp_19_validation",
                 train_labels_csv="wi_comp_19val_ground_truth.csv",
                 test_saples_folder="wi_comp_19_test",
                 test_labels_csv="wi_comp19_test_ground_truth.csv",
                 validation=False, split_images=[1, 1], transform=None) -> None:
        super().__init__()
        if transform is None:
            transform = transform_test
        if validation:
            csv_path = f"{ds_root}/{train_labels_csv}"
            image_folder = train_saples_folder
        else:
            csv_path = f"{ds_root}/{test_labels_csv}"
            image_folder = test_saples_folder
        self.transform = transform
        self.ids = []
        self.labels = []
        self.img_split = split_images
        self.split_count = split_images[0] * split_images[1]
        for line in open(csv_path, "r").readlines():
            line = line.strip()
            if line != "" and line[:4] != "FILE":
                fname, label = line.split(",")[:2]
                for _ in range(self.split_count):
                    self.ids.append(fname)
                    self.labels.append(int(label))
        self.labels = np.array(self.labels)
        self.ids = np.array(self.ids)
        self.root = ds_root
        self.image_path_pattern = f"{ds_root}/{image_folder}/{{}}"

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int, str]:
        img = Image.open(self.image_path_pattern.format(self.ids[index]))
        img = self.transform(img)
        if self.split_count > 1:
            partition = index % self.split_count
            h_partition = partition % self.img_split[0]
            v_partion = partition // self.img_split[0]
            left = h_partition * img.size(2) // self.img_split[0]
            right = (h_partition + 1) * img.size(2) // self.img_split[0]
            top = v_partion * img.size(1) // self.img_split[1]
            bottom = (v_partion + 1) * img.size(1) // self.img_split[1]
            img = img[:, top:bottom, left:right]
        return img, self.labels[index], self.ids[index]

    def __len__(self) -> int:
        return len(self.ids)

    def get_ids(self) -> np.array:
        return self.ids.copy()

    def get_labels(self) -> np.array:
        return self.labels.copy()


class WI2013(TextureDs):
    """Contemporary Writer identification dataset


    Citation: http://users.iit.demokritos.gr/~bgat/ICDAR_2013_Louloudis.pdf
    """
    urls = [
        'http://users.iit.demokritos.gr/~louloud/ICDAR2013WriterIdentificationComp/experimental_dataset_2013.rar',
        'http://users.iit.demokritos.gr/~louloud/ICDAR2013WriterIdentificationComp/icdar2013_benchmarking_dataset.rar'
    ]
    raw_folder = 'raw'
    processed_folder = 'processed'
    training_file = 'training.pt'
    test_file = 'test.pt'

    def __init__(self, root="tmp/wi2013", train=True, transform=transform_test, target_transform=None, download=False,
                 output_class="writer"):
        """Instanciates a pytorch dataset.

        :param root: The directory were everything is stored.
        :param train: Boolean, selects the data-partion and the default preprocessing
        :param transform: Image preprocessing use wi2013.transform_384_train for images of 512x192 obtained by cropping
                randomly and than scaling by 0.5, or wi2013.transform_test for variable sized images.
        :param target_transform: Ground truth preprocessing
        :param download: must be true for the data to download if missing
        :param output_class: This dataset supports different classes "writer":0-99,"sample":0-3 ,"script":0-1
        """
        assert output_class in ("writer", "sample", "script")
        if transform is None:
            if train:
                transform = transform_384_train
            else:
                transform = transform_test
        self.output_select = {"writer": lambda x: (x[0], x[1][0]-1, x[2]),
                              "sample": lambda x: (x[0], x[1][1]-1, x[2]),
                              "script": lambda x: (x[0], x[1][0]-1, x[2])}[output_class]
        self.root = os.path.expanduser(root)
        self.transform = transform
        self.target_transform = target_transform
        self.train = train  # training set or test set

        if download:
            self.download()

        if not self._check_exists():
            raise RuntimeError('Dataset not found.' +
                               ' You can use download=True to download it')

        if self.train:
            self.train_samples = torch.load(
                os.path.join(self.root, self.processed_folder, self.training_file))
        else:
            self.test_samples = torch.load(
                os.path.join(self.root, self.processed_folder, self.test_file))
        if self.train:
            items = self.train_samples
        else:
            items = self.test_samples
        print(f"Loading data: {len(items[0])}")
        img_label_id = [self.output_select(item) for item in items]
        self.labels = np.array([label for _, label, _ in img_label_id])
        self.ids = np.array([id for _, _, id in img_label_id])
        # sorting samples by filename
        id_idx = np.argsort(self.ids)
        self.labels = self.labels[id_idx]
        self.ids = self.ids[id_idx]
        if self.train:
            self.train_samples = [self.train_samples[idx] for idx in id_idx]
        else:
            self.test_samples = [self.test_samples[idx] for idx in id_idx]

    def __getitem__(self, index) -> Tuple[torch.Tensor, int, str]:
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        if self.train:
            img, label, id = self.output_select(self.train_samples[index])
            label = label - 26
        else:
            img, label, id = self.output_select(self.test_samples[index])

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = img  # .fromarray(img.numpy(), mode='L')

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            label = self.target_transform(label)

        return img, label, id

    def __len__(self):
        if self.train:
            return len(self.train_samples)
        else:
            return len(self.test_samples)

    def get_labels(self) -> np.array:
        return self.labels

    def get_ids(self) -> np.array:
        return self.ids

    def _check_exists(self):
        return os.path.exists(os.path.join(self.root, self.processed_folder, self.training_file)) and \
            os.path.exists(os.path.join(self.root, self.processed_folder, self.test_file))

    def download(self):
        """Download the rar files data if it doesn't exist in processed_folder already."""
        import urllib
        import rarfile

        if self._check_exists():
            return

        # download files
        try:
            os.makedirs(os.path.join(self.root, self.raw_folder))
            os.makedirs(os.path.join(self.root, self.processed_folder))
        except OSError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise

        for url in self.urls:
            print('Downloading ' + url)
            stream = urllib.request.urlopen(url)
            filename = url.rpartition('/')[-1]
            file_path = os.path.join(self.root, self.raw_folder, filename)
            with open(file_path, 'wb') as f:
                f.write(stream.read())
            with rarfile.RarFile(file_path) as rar_f:
                rar_f.extractall(self.raw_folder)
            # os.unlink(file_path)

        # process and save as torch files
        print('Processing...')
        train_set = read_rar_file(rarfile.RarFile(os.path.join(self.root, self.raw_folder,
                                                               'experimental_dataset_2013.rar')))
        test_set = read_rar_file(
            rarfile.RarFile(os.path.join(self.root, self.raw_folder, 'icdar2013_benchmarking_dataset.rar')))
        with open(os.path.join(self.root, self.processed_folder, self.training_file), 'wb') as f:
            torch.save(train_set, f)
        with open(os.path.join(self.root, self.processed_folder, self.test_file), 'wb') as f:
            torch.save(test_set, f)
        print('Done!')

    def __repr__(self):
        fmt_str = 'Dataset ' + self.__class__.__name__ + '\n'
        fmt_str += '    Number of datapoints: {}\n'.format(self.__len__())
        tmp = 'train' if self.train is True else 'test'
        fmt_str += '    Split: {}\n'.format(tmp)
        fmt_str += '    Root Location: {}\n'.format(self.root)
        tmp = '    Transforms (if any): '
        fmt_str += '{0}{1}\n'.format(tmp, self.transform.__repr__().replace('\n', '\n' + ' ' * len(tmp)))
        tmp = '    Target Transforms (if any): '
        fmt_str += '{0}{1}'.format(tmp, self.target_transform.__repr__().replace('\n', '\n' + ' ' * len(tmp)))
        return fmt_str


hisfrag_transform = transforms.Compose([
    transforms.Grayscale(),
    transforms.ToTensor(),
])


class HisFrag20():
    def __init__(self, img_dir, transform=hisfrag_transform, return_writers=True):
        self.image_paths = list(sorted(glob.glob(f"{img_dir}/*.jpg")))
        if len(self.image_paths) == 0:
            raise FileNotFoundError(f"No images found in {img_dir}")
        self.transform = transform
        if return_writers:
            idx = 0
        else:
            idx = 1
        self.labels = np.array([int(Path(img).stem.split('_')[idx]) for img in self.image_paths])
        self.ids = np.array([img_path for img_path in self.image_paths])
        self.unique_labels, self.unique_labels_count = np.unique(self.labels, return_counts=True)
        self.by_label = defaultdict(lambda: [])
        for label, path in zip(self.labels, self.image_paths):
            self.by_label[label].append(path)

    def __getitem__(self, index: int) -> tuple[torch.Tensor, int, str]:
        img = Image.open(self.image_paths[index])
        return self.transform(img), self.labels[index], self.image_paths[index]

    def __len__(self):
        return len(self.image_paths)

    def get_ids(self):
        return self.ids

    def get_labels(self):
        return self.labels


def _warn(*args):
    sys.stderr.write(" ".join([str(arg) for arg in args])+"\n")


def _get_dict(compressed_stream, filter_gt=False, filter_nongt=False):
    assert not (filter_gt and filter_nongt)

    def isimage(x) -> bool:
        return x.split(".")[-1].lower() in ["tiff", "bmp", "jpg", "tif", "jpeg", "png"] and "skel" not in x.lower()

    def isgt(x) -> bool:
        return isimage(x) and ("gt" in x or "GT" in x)

    if isinstance(compressed_stream, py7zr.SevenZipFile):
        compressed_stream.reset()
        res_dict = compressed_stream.readall()
        names = res_dict.keys()
        if filter_gt:
            names = [n for n in names if not isgt(n)]
        if filter_nongt:
            names = [n for n in names if isgt(n)]
        return {name: res_dict[name] for name in names}
    elif isinstance(compressed_stream, rarfile.RarFile) or isinstance(compressed_stream, zipfile.ZipFile):
        names = compressed_stream.namelist()
        names = [n for n in names if isimage(n)]
        if filter_gt:
            names = [n for n in names if not isgt(n)]
        if filter_nongt:
            names = [n for n in names if isgt(n)]
        return {name: BytesIO(compressed_stream.read(compressed_stream.getinfo(name))) for name in names}
    else:
        raise ValueError("Filedescriptor must be one of [rar, zip, 7z]")


def _extract(archive, root=None) -> None:
    if archive.endswith(".tar.gz"):
        if root is None:
            cmd = "tar -xpvzf {}".format(archive)
        else:
            cmd = 'mkdir -p {};tar -xpvzf {} -C{}'.format(root, archive, root)
        _ = shell_stdout(cmd)
    else:
        raise NotImplementedError()


def _check_os_dependencies() -> bool:
    program_list = ["wget"]
    return all([shell_stdout("which "+prog) for prog in program_list])


def _mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def _resumable_download(url, save_dir):
    _mkdir_p(save_dir)
    download_cmd = 'wget --directory-prefix=%s -c %s' % (save_dir, url)
    _warn("Downloading {} ... ".format(url))
    shell_stdout(download_cmd)
    _warn("done")
    return os.path.join(save_dir, url.split("/")[-1])


dibco_transform_gray_input = torchvision.transforms.Compose([
    torchvision.transforms.Grayscale(),
    torchvision.transforms.ToTensor(),
])

dibco_transform_color_input = torchvision.transforms.Compose([
    torchvision.transforms.ToTensor(),
])

dibco_transform_gt = torchvision.transforms.Compose([
    torchvision.transforms.Grayscale(),
    torchvision.transforms.ToTensor(),
])


class Dibco:
    """Provides one or more of the `DIBCO <https://vc.ee.duth.gr/dibco2019/>` datasets.

    Os dependencies: Other than python packages, unrar and arepack CLI tools must be installed.
    In Ubuntu they can be installed with: sudo apt install unrar atool p7zip-full
    In order to concatenate two DIBCO datasets just add them:
    .. source :: python

        trainset = dibco.Dibco.Dibco2009() + dibco.Dibco.Dibco2013()
        valset = dibco.Dibco.Dibco2017() + dibco.Dibco.Dibco209()

    Each item is a tuple of an RGB PIL image and an Binary PIL image. The images are transformed by ``input_transform``
    and ``gt_transform``.
    """
    urls = {
        "2009_HW": ["https://users.iit.demokritos.gr/~bgat/DIBCO2009/benchmark/DIBC02009_Test_images-handwritten.rar",
                    "https://users.iit.demokritos.gr/~bgat/DIBCO2009/benchmark/DIBCO2009-GT-"
                    "Test-images_handwritten.rar"],

        "2009_P": ["https://users.iit.demokritos.gr/~bgat/DIBCO2009/benchmark/DIBCO2009_Test_images-printed.rar",
                   "https://users.iit.demokritos.gr/~bgat/DIBCO2009/benchmark/DIBCO2009-GT-Test-images_printed.rar"],

        "2010": ["http://users.iit.demokritos.gr/~bgat/H-DIBCO2010/benchmark/H_DIBCO2010_test_images.rar",
                 "http://users.iit.demokritos.gr/~bgat/H-DIBCO2010/benchmark/H_DIBCO2010_GT.rar"],

        "2011_P": ["http://utopia.duth.gr/~ipratika/DIBCO2011/benchmark/dataset/DIBCO11-machine_printed.rar"],
        "2011_HW": ["http://utopia.duth.gr/~ipratika/DIBCO2011/benchmark/dataset/DIBCO11-handwritten.rar"],

        "2012": ["http://utopia.duth.gr/~ipratika/HDIBCO2012/benchmark/dataset/H-DIBCO2012-dataset.rar"],

        "2013": ["http://utopia.duth.gr/~ipratika/DIBCO2013/benchmark/dataset/DIBCO2013-dataset.rar"],

        "2014": ["http://users.iit.demokritos.gr/~bgat/HDIBCO2014/benchmark/dataset/original_images.rar",
                 "http://users.iit.demokritos.gr/~bgat/HDIBCO2014/benchmark/dataset/GT.rar"],
        "2016": ["https://vc.ee.duth.gr/h-dibco2016/benchmark/DIBCO2016_dataset-original.zip",
                 "https://vc.ee.duth.gr/h-dibco2016/benchmark/DIBCO2016_dataset-GT.zip"],
        "2017": ["https://vc.ee.duth.gr/dibco2017/benchmark/DIBCO2017_Dataset.7z",
                 "https://vc.ee.duth.gr/dibco2017/benchmark/DIBCO2017_GT.7z"],
        "2018": ["http://vc.ee.duth.gr/h-dibco2018/benchmark/dibco2018_Dataset.zip",
                 "http://vc.ee.duth.gr/h-dibco2018/benchmark/dibco2018-GT.zip"],
        "2019A": ["https://vc.ee.duth.gr/dibco2019/benchmark/dibco2019_dataset_trackA.zip",
                  "https://vc.ee.duth.gr/dibco2019/benchmark/dibco2019_gt_trackA.zip"],
        "2019B": ["https://vc.ee.duth.gr/dibco2019/benchmark/dibco2019_dataset_trackB.zip",
                  "https://vc.ee.duth.gr/dibco2019/benchmark/dibco2019_GT_trackB.zip"]
    }

    # urls = {
    #     "2009_HW": ["http://rr.visioner.ca/assets/dibco_mirror/DIBC02009_Test_images-handwritten.rar",
    #                 "http://rr.visioner.ca/assets/dibco_mirror/DIBCO2009-GT-Test-images_handwritten.rar"],
    #     "2009_P": ["http://rr.visioner.ca/assets/dibco_mirror/DIBCO2009_Test_images-printed.rar",
    #                "http://rr.visioner.ca/assets/dibco_mirror/DIBCO2009-GT-Test-images_printed.rar"],
    #     "2010": ["http://rr.visioner.ca/assets/dibco_mirror/H_DIBCO2010_test_images.rar",
    #              "http://rr.visioner.ca/assets/dibco_mirror/H_DIBCO2010_GT.rar"],
    #     "2011_P": ["http://rr.visioner.ca/assets/dibco_mirror/DIBCO11-machine_printed.rar"],
    #     "2011_HW": ["http://rr.visioner.ca/assets/dibco_mirror/DIBCO11-handwritten.rar"],
    #     "2012": ["http://rr.visioner.ca/assets/dibco_mirror/H-DIBCO2012-dataset.rar"],
    #     "2013": ["http://rr.visioner.ca/assets/dibco_mirror/DIBCO2013-dataset.rar"],
    #     "2014": ["http://rr.visioner.ca/assets/dibco_mirror/original_images.rar",
    #              "http://rr.visioner.ca/assets/dibco_mirror/GT.rar"],
    #     "2016": ["http://rr.visioner.ca/assets/dibco_mirror/DIBCO2016_dataset-original.zip",
    #              "http://rr.visioner.ca/assets/dibco_mirror/DIBCO2016_dataset-GT.zip"],
    #     "2017": ["http://rr.visioner.ca/assets/dibco_mirror/DIBCO2017_Dataset.7z",
    #              "http://rr.visioner.ca/assets/dibco_mirror/DIBCO2017_GT.7z"],
    #     "2018": ["http://rr.visioner.ca/assets/dibco_mirror/dibco2018_Dataset.zip",
    #              "http://rr.visioner.ca/assets/dibco_mirror/dibco2018-GT.zip"],
    #     "2019A": ["http://rr.visioner.ca/assets/dibco_mirror/dibco2019_dataset_trackA.zip",
    #               "http://rr.visioner.ca/assets/dibco_mirror/dibco2019_gt_trackA.zip"],
    #     "2019B": ["http://rr.visioner.ca/assets/dibco_mirror/dibco2019_dataset_trackB.zip",
    #               "http://rr.visioner.ca/assets/dibco_mirror/dibco2019_GT_trackB.zip"]
    # }

    @staticmethod
    def load_single_stream(compressed_stream):
        inname2bs = _get_dict(compressed_stream, filter_gt=True)
        gtname2bs = _get_dict(compressed_stream, filter_nongt=True)
        id2gt = {n.split("/")[-1].split("_")[0].split(".")[0]: Image.open(fd).copy() for n, fd in gtname2bs.items()}
        id2in = {n.split("/")[-1].split("_")[0].split(".")[0]: Image.open(fd).copy() for n, fd in inname2bs.items()}
        assert set(id2gt.keys()) == set(id2in.keys())
        id2in = {k: v.convert("RGB") for k, v in id2in.items()}
        return {k: (id2in[k], id2gt[k]) for k in id2gt.keys()}

    @staticmethod
    def load_double_stream(input_compressed_stream, gt_compressed_stream):
        inname2bs = _get_dict(input_compressed_stream)
        gtname2bs = _get_dict(gt_compressed_stream)
        id2in = {n.split("/")[-1].split("_")[0].split(".")[0]: Image.open(fd).copy() for n, fd in inname2bs.items()}
        id2gt = {n.split("/")[-1].split("_")[0].split(".")[0]: Image.open(fd).copy() for n, fd in gtname2bs.items()}
        assert set(id2gt.keys()) == set(id2in.keys())
        id2in = {k: v.convert("RGB") for k, v in id2in.items()}
        return {k: (id2in[k], id2gt[k]) for k in id2gt.keys()}

    @staticmethod
    def Dibco2009(**kwargs):
        kwargs["partitions"] = ["2009_HW", "2009_P"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2010(**kwargs):
        kwargs["partitions"] = ["2010"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2011(**kwargs):
        kwargs["partitions"] = ["2011_P", "2011_HW"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2012(**kwargs):
        kwargs["partitions"] = ["2012"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2013(**kwargs):
        kwargs["partitions"] = ["2013"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2014(**kwargs):
        kwargs["partitions"] = ["2014"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2016(**kwargs):
        kwargs["partitions"] = ["2016"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2017(**kwargs):
        kwargs["partitions"] = ["2017"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2018(**kwargs):
        kwargs["partitions"] = ["2018"]
        return Dibco(**kwargs)

    @staticmethod
    def Dibco2019(**kwargs):
        kwargs["partitions"] = ["2019A", "2019B"]
        return Dibco(**kwargs)

    def __init__(self, partitions=["2009_HW", "2009_P"], root="./tmp/dibco",
                 input_transform=dibco_transform_gray_input, gt_transform=dibco_transform_gt, add_mask=False):
        self.input_transform = input_transform
        self.gt_transform = gt_transform
        self.root = root
        self.add_mask = add_mask
        data = {}
        for partition in partitions:
            for url in Dibco.urls[partition]:
                archive_fname = root + "/" + url.split("/")[-1]
                if not os.path.isfile(archive_fname):
                    _resumable_download(url, root)
                else:
                    _warn(archive_fname, " found in cache.")
            if len(Dibco.urls[partition]) == 2:
                if Dibco.urls[partition][0].endswith(".rar"):
                    input_rar = rarfile.RarFile(root + "/" + Dibco.urls[partition][0].split("/")[-1])
                    gt_rar = rarfile.RarFile(root + "/" + Dibco.urls[partition][1].split("/")[-1])
                    samples = {partition + "/" + k: v for k, v in Dibco.load_double_stream(input_rar, gt_rar).items()}
                    data.update(samples)
                elif Dibco.urls[partition][0].endswith(".zip") or Dibco.urls[partition][0].endswith(".7z"):
                    zip_input_fname = root + "/" + Dibco.urls[partition][0].split("/")[-1]
                    zip_gt_fname = root + "/" + Dibco.urls[partition][1].split("/")[-1]
                    if zip_input_fname.endswith("7z"):
                        input_zip = py7zr.SevenZipFile(zip_input_fname)
                    else:
                        input_zip = zipfile.ZipFile(zip_input_fname)
                    if zip_gt_fname.endswith("7z"):
                        gt_zip = py7zr.SevenZipFile(zip_gt_fname)
                    else:
                        gt_zip = zipfile.ZipFile(zip_gt_fname)
                    samples = {partition + "/" + k: v for k, v in Dibco.load_double_stream(input_zip, gt_zip).items()}
                    data.update(samples)
                else:
                    raise ValueError("Unknown file type")
            else:
                if Dibco.urls[partition][0].endswith(".rar"):
                    input_rar = rarfile.RarFile(root + "/" + Dibco.urls[partition][0].split("/")[-1])
                    samples = {partition + "/" + k: v for k, v in Dibco.load_single_stream(input_rar).items()}
                    data.update(samples)
                elif Dibco.urls[partition][0].endswith(".zip") or Dibco.urls[partition][0].endswith(".7z"):
                    zip_input_fname = root + "/" + Dibco.urls[partition][0].split("/")[-1]
                    if zip_input_fname.endswith("7z"):
                        # zip_input_fname = zip_input_fname[:-2] + "zip"
                        input_zip = py7zr.SevenZipFile(zip_input_fname)
                    else:
                        input_zip = zipfile.ZipFile(zip_input_fname)
                    samples = {partition + "/" + k: v for k, v in Dibco.load_single_stream(input_zip).items()}
                    data.update(samples)
                else:
                    raise ValueError("Unknown file type")
        id_data = list(data.items())
        self.sample_ids = [sample[0] for sample in id_data]
        self.inputs = [sample[1][0] for sample in id_data]
        self.gt = [sample[1][1] for sample in id_data]

    def __getitem__(self, item):
        input_img = self.input_transform(self.inputs[item])
        gt = self.gt_transform(self.gt[item])
        if self.add_mask:
            return input_img, gt, torch.ones_like(input_img[:1, :, :])
        else:
            return input_img, gt

    def __len__(self):
        return len(self.sample_ids)

    def __add__(self, other):
        res = Dibco(partitions=[])
        res.root = self.root
        res.input_transform = self.input_transform
        res.gt_transform = self.gt_transform
        res.sample_ids = self.sample_ids + other.sample_ids
        res.inputs = self.inputs + other.inputs
        res.gt = self.gt + other.gt
        return res


all_dibco_keys = set(Dibco.urls.keys())


l1out_partitions = {}
for k in all_dibco_keys:
    trainset = sorted(all_dibco_keys - set([k]))
    trainset, validationset = trainset[:-1], trainset[-1:]
    l1out_partitions[k] = {"train": trainset, "val": validationset, "test": k}



import pytest
import torch
from ptlbp import PointSampler, Comparisson, ComparissonOtsu, ComparissonQuantile
import ptlbp
from ptlbp.testing import generate_rect_image, generate_noise_image, generate_bimodal_image


print_plot_debug = False


@pytest.mark.parametrize(["img", "pointsampler", "comparisson", "true_count_range"], [

    (generate_rect_image(100), PointSampler(1, "8r1"), Comparisson(), (580, 600)),
    (generate_rect_image(100), PointSampler(1, "8r1"), ComparissonOtsu(), (580, 600)),
    (generate_rect_image(100), PointSampler(1, "8r1"), ComparissonQuantile(), (580, 600)),

    (torch.zeros(1, 1, 100, 100), PointSampler(1, "8r1"), Comparisson(), (0, 0)),
    (torch.ones(1, 1, 100, 100), PointSampler(1, "8r1"), Comparisson(), (0, 0)),

    (torch.zeros(1, 1, 100, 100), PointSampler(1, "8r1"), ComparissonOtsu(), (0, 0)),
    (torch.ones(1, 1, 100, 100), PointSampler(1, "8r1"), ComparissonOtsu(), (0, 0)),

    (torch.zeros(1, 1, 100, 100), PointSampler(1, "8r1"), ComparissonQuantile(), (0, 0)),
    (torch.ones(1, 1, 100, 100), PointSampler(1, "8r1"), ComparissonQuantile(), (0, 0)),

    (generate_noise_image(100), PointSampler(1, "8r1"), Comparisson(), (19000, 19100)),
    (generate_noise_image(100), PointSampler(1, "8r1"), ComparissonOtsu(), (560, 600)),
    (generate_noise_image(100), PointSampler(1, "8r1"), ComparissonQuantile(), (15200, 15300)),

    (generate_bimodal_image(100), PointSampler(1, "8r1"), Comparisson(), (9800, 10200)),
    (generate_bimodal_image(100), PointSampler(1, "8r1"), ComparissonOtsu(), (400, 600)),
    (generate_bimodal_image(100), PointSampler(1, "8r1"), ComparissonQuantile(), (6800, 7200)),

    (1 - generate_bimodal_image(100), PointSampler(1, "8r1"), Comparisson(), (11200, 12200)),
    (1 - generate_bimodal_image(100), PointSampler(1, "8r1"), ComparissonOtsu(), (12200, 12300)),
    (1 - generate_bimodal_image(100), PointSampler(1, "8r1"), ComparissonQuantile(), (8350, 8750)),
    ][-3:])
def test_comparisson(img, pointsampler: PointSampler, comparisson: torch.nn.Module, true_count_range: tuple):
    torch.manual_seed(1337)
    x = img
    offsets = pointsampler(x)
    soft_output = comparisson(x, offsets)
    harder_output = torch.tanh(100. * soft_output) * .5 + .5
    hard_output = (harder_output > .5).float()
    if print_plot_debug:
        ptlbp.util.plot_img(img, title=f"Input {type(comparisson)}", block=False)
        ptlbp.util.plot_img(soft_output[:, :1, :, :], title=f"Slice 0, Soft {type(comparisson)}", block=False)
        ptlbp.util.plot_img(harder_output[:, :1, :, :], title=f"Slice 0, Harder {type(comparisson)}", block=True)
        print(comparisson, harder_output.sum(), harder_output.mean())
    assert true_count_range[0] <= hard_output.sum().item() <= true_count_range[1]


def test_zeros():
    x = torch.zeros(1, 1, 100, 100)
    pointsampler = PointSampler(1, "8r1")
    comparisson = ComparissonOtsu()
    offsets = pointsampler(x)
    comparissons = comparisson(x, offsets)
    assert (comparissons < 0).float().mean().item() == 1.



from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec
from PIL import Image
import numpy as np
import torch
import matplotlib
from contextlib import nullcontext
import sklearn
import tqdm


def conditional_no_grad(condition):
    if condition:
        return torch.no_grad()
    else:
        return nullcontext()


def sk_mAP(X, y, metric='l1', progress_bar_freq=-1):
    non_singleton_idx = (y[None, :] == y[:, None]).sum(axis=1) > 1
    X, y = X[non_singleton_idx, :], y[non_singleton_idx]
    num_samples = X.shape[0]
    num_estimated_samples = 0
    map_score = 0.0
    top1 = []
    if progress_bar_freq > 0:
        range_iter = tqdm.tqdm(range(num_samples), total=num_samples,
                               desc='mAP', position=0, leave=True, mininterval=1)
    else:
        range_iter = range(num_samples)
    for i in range_iter:
        y_true = (y == y[i]).astype(int)  # Binary relevance: 1 if same class, 0 otherwise
        if metric == 'l1':
            y_score = np.sum(np.abs(X - X[i]), axis=1)  # L1 distances between sample i and all other samples
        elif metric == 'l2':
            y_score = np.sum((X - X[i])**2, axis=1)
        else:
            raise NotImplementedError(f"Metric {metric} not implemented")
        exclude_self = np.arange(num_samples) != i
        y_true, y_score = y_true[exclude_self], y_score[exclude_self]
        if y_true.sum() > 0:
            top1.append(y_true[y_score.argmin()])
            # Negative y_score as sklearn expects higher scores to be better
            map_score += sklearn.metrics.average_precision_score(y_true, -y_score)
            num_estimated_samples += 1
    map_score /= num_estimated_samples
    return map_score, np.array(top1).mean()


def mAP(X, y, metric='l1'):
    gt_table = (y[None, :] == y[:, None]).astype(int)
    dm = sklearn.metrics.pairwise_distances(X, Y=None, metric='euclidean', n_jobs=None, force_all_finite=True)


def pil_image_as_batch(img, add_noise=0, add_large_gradient=0, device='cpu'):
    img = img.convert('L')
    img = np.array(img).astype(float)/255.
    img = img[None, None, :, :]
    res_img = torch.tensor(img).float().to(device)
    if add_noise > 0:
        N = torch.rand_like(res_img)
        res_img = res_img * (1-add_noise) + add_noise * N
    if add_large_gradient > 0:
        a = torch.ones_like(res_img)
        a = torch.cumsum(a, dim=3)
        a = a/a.max()
        res_img = res_img * (1-add_large_gradient) + add_large_gradient * a
    return res_img


def np_image_as_batch(img, add_noise=0, add_large_gradient=0, device='cpu'):
    img = img.astype(float)/255.
    if img.ndim == 2:
        img = img[None, None, :, :]
    elif img.ndim == 3:
        img = img.transpose(2, 0, 1)
        img = img[None, :, :, :]
    res_img = torch.tensor(img).float().to(device)
    if add_noise > 0:
        N = torch.rand_like(res_img)
        res_img = res_img * (1-add_noise) + add_noise * N
    if add_large_gradient > 0:
        a = torch.ones_like(res_img)
        a = torch.cumsum(a, dim=3)
        a = a/a.max()
        res_img = res_img * (1-add_large_gradient) + add_large_gradient * a
    return res_img


def load_image_as_batch(path, add_noise=0, add_large_gradient=0, device='cpu'):
    img = Image.open(path)
    return pil_image_as_batch(img, add_noise=add_noise, add_large_gradient=add_large_gradient, device=device)


def plot_lbp(img, title='', fig=None, block=False):
    if fig is None:
        fig = plt.figure(constrained_layout=False)
    lab_img = img.detach().cpu().numpy()[0, :, :, :].argmax(axis=0)
    plt.imshow(lab_img, cmap='d256')
    plt.xticks([], [])
    if title:
        plt.title(title)
    plt.show(block=block)
    return fig


def plot_img(img, title='', cmap='gray', fig=None, block=False):
    if fig is None:
        fig = plt.figure(constrained_layout=False)
    np_img = img.detach().cpu().numpy()[0, 0, :, :]
    plt.imshow(np_img, cmap='gray')
    plt.colorbar()
    plt.xticks([], [])
    if title:
        plt.title(title)
    plt.show(block=block)


def plot_slices(img, title='', cols=4, cmap='gray', block=False):
    val_min = img.min().item()
    val_max = img.max().item()
    rows = (img.size(1) // cols) + int((img.size(1) % cols) > 0)
    fig = plt.figure(constrained_layout=False)
    widths = (1 * np.ones(cols)).tolist() + [.05]
    heights = (1 * np.ones(rows)).tolist()
    gs = GridSpec(rows, cols+1, figure=fig, width_ratios=widths,
                  height_ratios=heights, left=0.00, right=1., wspace=0.05, hspace=0.05)
    cbar_ax = fig.add_subplot(gs[:, -1])
    for slice in range(img.shape[1]):
        row = slice // cols
        col = slice % cols
        np_img = img.detach().cpu().numpy()[0, slice, :, :]
        ax = fig.add_subplot(gs[row, col])
        imax = ax.imshow(np_img, cmap='gray', extent=[0, 100, 0, 100], vmin=val_min, vmax=val_max)
        ax.axes.xaxis.set_ticks([])
        ax.axes.yaxis.set_ticks([])
        if slice == img.size(1)-1:
            fig.colorbar(imax, cax=cbar_ax)
            pass
    if title:
        fig.suptitle(title)
    plt.show(block=block)


def create_dicrete_colormap(N, pattern0=0):
    patternValues = np.cumsum(np.ones((N + 1), dtype='uint32')) - 1
    resFloatRed = 0.45 * np.mod(patternValues, 2) + \
        0.325 * np.mod(patternValues / 8, 2) + 0.225 * np.mod(patternValues / 64, 2)
    resFloatGreen = 0.45 * np.mod(patternValues / 2, 2) + \
        0.325 * np.mod(patternValues / 16, 2) + 0.225 * np.mod(patternValues / 128, 2)
    resFloatBlue = 0.6 * np.mod(patternValues / 4, 2) + 0.4 * np.mod(patternValues / 32, 2)
    x = np.linspace(0, 1, N)
    rXY0Y1 = np.concatenate((x.reshape(-1, 1), resFloatRed[:-1].reshape(-1, 1),
                             resFloatRed[:-1].reshape(-1, 1)), axis=1)
    r = tuple([tuple(line) for line in list(rXY0Y1)])
    gXY0Y1 = np.concatenate((x.reshape(-1, 1), resFloatGreen[:-1].reshape(-1, 1),
                             resFloatGreen[:-1].reshape(-1, 1)), axis=1)
    g = tuple([tuple(line) for line in gXY0Y1])
    bXY0Y1 = np.concatenate((x.reshape(-1, 1), resFloatBlue[:-1].reshape(-1, 1),
                             resFloatBlue[:-1].reshape(-1, 1)), axis=1)
    b = tuple([tuple(line) for line in bXY0Y1])
    cDict = {'red': r, 'green': g, 'blue': b}
    cMap = matplotlib.colors.LinearSegmentedColormap(f"d{N}", cDict)
    matplotlib.colormaps.register(cMap)



import torch
import numpy as np
from typing import Tuple


class BinaryToOnehotChannelEmbeddings(torch.nn.Module):
    @staticmethod
    def img2embedding(img_BCHW: torch.Tensor) -> Tuple[Tuple[int, int, int, int], torch.Tensor]:
        original_sizes = tuple(img_BCHW.size())
        img_BWHC = img_BCHW.swapaxes(1, 3)
        img_BWH_C = img_BWHC.reshape(-1, original_sizes[1])
        return original_sizes, img_BWH_C

    @staticmethod
    def embedding2img(original_sizes: Tuple[int, int, int, int], img_BWH_C: torch.Tensor) -> torch.Tensor:
        b_sz, c_sz, h_sz, w_sz = original_sizes
        img_BWHC = img_BWH_C.view(b_sz, w_sz, h_sz, c_sz)
        img_BCHW = img_BWHC.swapaxes(1, 3)
        return img_BCHW

    @staticmethod
    def powerset_map(input_length: int) -> list[int]:
        powerset = []
        for n in range(2**input_length):
            powerset.append([(n//(2**k)) % 2 for k in range(0, input_length)])
        return powerset

    def __init__(self, in_channels) -> None:
        super().__init__()
        W = np.array(BinaryToOnehotChannelEmbeddings.powerset_map(in_channels))[:, :, None, None]
        self.linear = torch.nn.Linear(in_channels, 2 ** in_channels, bias=False)
        self.linear.weight.data = torch.tensor(W).float()[:, :, 0, 0]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_sz, x = BinaryToOnehotChannelEmbeddings.img2embedding(x)  # x == x_BC
        margin = -x.sum(dim=1)[:, None] + 1
        x = margin + self.linear(x) - self.linear(1-x)
        res_sz = x_sz[0], 2**x_sz[1], x_sz[2], x_sz[3]
        x = BinaryToOnehotChannelEmbeddings.embedding2img(res_sz, x)
        return x

    @property
    def num_output_channels(self) -> int:
        return self.linear.out_features


# class BinaryToOnehotChannelConvolutions(torch.nn.Module):
#     @staticmethod
#     def powerset_map(input_length) -> list:
#         powerset = []
#         for n in range(2**input_length):
#             powerset.append([(n//(2**k)) % 2 for k in range(0, input_length)])
#         return powerset

#     def __init__(self, in_channels, output_hardness=3.0, input_hardness=3.0,
#                  learn_input_hardness=False, learn_output_hardness=False):
#         super().__init__()
#         W = np.array(BinaryToOnehotChannelConvolutions.powerset_map(in_channels))[:, :, None, None]
#         self.weight = torch.nn.parameter.Parameter(torch.tensor(W, dtype=torch.float32))

#     def forward(self, x):
#         positive = torch.nn.functional.conv2d(x, self.weight)
#         negative = torch.nn.functional.conv2d(1-x, self.weight)
#         result = 1 + positive - negative
#         return result



from typing import Tuple, Union
import torch
from .diff_lbp import DiffLBP
from .util import conditional_no_grad
from sklearn.decomposition import PCA
import sys
from pathlib import Path
import tqdm


def hellinger_normalise(features):
    #  return torch.sign(features) * torch.abs(features)**.5
    abs_features = (features ** 2) ** .5
    sign = features / (abs_features + 1e-10)
    return sign * (abs_features ** .5)


def l2_normalise(features, epsilon=1e-10):
    return features/(((torch.sum(features**2, dim=1)**.5)+epsilon)[:, None])


# class DiffSRSLBP(torch.nn.Module):
#     def __init__(self, num_components=200, radii=range(1, 12), supress_zero_pattern=True, supress_full_pattern=True,
#                  block_normalise=True, gradient_on_lbp=True, gradient_on_srs=True, apply_helliger=True, apply_l2=True):
#         super().__init__()
#         radii_modules = []
#         for r in radii:
#             radii_modules.append(DiffLBP(f"8r{r}", supress_zero_pattern=supress_zero_pattern,
#                                          supress_full_pattern=supress_full_pattern,
#                                          block_normalise=block_normalise, comparisson="simple"))
#         self.diff_lbp = torch.nn.ModuleList(radii_modules)
#         if num_components <= 0:
#             self.feature_reduction = None
#         else:
#             self.feature_reduction = torch.nn.Linear(self.histograms_size, num_components, bias=False)

#         # We cant have gradient only on early layers
#         assert (gradient_on_srs and gradient_on_lbp) or not gradient_on_srs
#         self.gradient_on_lbp = gradient_on_lbp
#         self.gradient_on_srs = gradient_on_srs

#         self.apply_helliger = apply_helliger
#         self.apply_l2 = apply_l2

#     def __forward_lbp(self, x):
#         with conditional_no_grad(not self.gradient_on_lbp):
#             x = torch.cat([diff_lbp(x) for diff_lbp in self.diff_lbp], dim=1)
#         return x

#     def __forward_features(self, x):
#         with conditional_no_grad(not self.gradient_on_srs):
#             x = x[:, :, 0, 0]
#             if self.feature_reduction is not None:
#                 x = self.feature_reduction(x)
#             x = x[:, :, None, None]
#             if self.apply_helliger:
#                 x = hellinger_normalise(x)
#             if self.apply_l2:
#                 x = l2_normalise(x)
#         return x

#     def forward(self, x):
#         x = self.__forward_lbp(x)
#         x = self.__forward_features(x)
#         return x

#     def train_pca(self, data):
#         assert self.feature_reduction
#         with torch.no_grad():
#             if isinstance(data, torch.utils.data.DataLoader):
#                 features = []
#                 for sample in data:
#                     features.append(torch.cat([diff_lbp(sample) for diff_lbp in self.diff_lbp], dim=1))
#                 block_normalised_features = torch.cat(features, dim=0)
#             elif isinstance(data, torch.Tensor) and data.dim() == 4:
#                 assert data.size(2) == data.size(3) == 1
#                 block_normalised_features = data[:, :, 0, 0]
#             elif isinstance(data, torch.Tensor) and data.dim() == 2:
#                 block_normalised_features = data
#             else:
#                 raise ValueError(f"Data type {type(data)} or size not supported")

#             device = block_normalised_features.device
#             pca = PCA(n_components=self.num_components)
#             if block_normalised_features.size(0) < self.num_components:
#                 replicates = self.num_components // block_normalised_features.size(0) + 1
#                 new_feature_list = []
#                 for _ in range(replicates):
#                     noise = torch.randn_like(block_normalised_features) * .0000001
#                     new_feature_list.append(block_normalised_features + noise)
#                 block_normalised_features = torch.cat(new_feature_list, dim=0)
#             assert block_normalised_features.size(2) == block_normalised_features.size(3) == 1
#             block_normalised_features = block_normalised_features[:, :, 0, 0]
#             pca.fit(block_normalised_features.cpu().numpy())
#             self.feature_reduction.weight.data = torch.tensor(pca.components_).to(device).float()

#     @property
#     def histograms_size(self):
#         return sum(diff_lbp.histogram_size for diff_lbp in self.diff_lbp)

#     @property
#     def num_components(self):
#         if self.feature_reduction is None:
#             return self.histograms_size
#         else:
#             return self.feature_reduction.out_features


class MultiRadiusDiffLBP(torch.nn.Module):
    def embbed_dataset(self, ds, resume_path, device=None, save_freq=1000, progress=True):
        if device is not None:
            self.to(device)
        embeders = self.diff_lbp
        features_labels_filenames = []
        processed_ids = {}
        if Path(resume_path).exists():
            try:
                features_labels_filenames = torch.load(resume_path)
                processed_ids = {id: (feats, label, id) for (feats, label, id) in features_labels_filenames}
                print(f"Resumed {len(processed_ids)} from {resume_path}", file=sys.stderr)
            except Exception:
                print(f"Error loading {resume_path}", file=sys.stderr)
                features_labels_filenames = []
                processed_ids = {}
        ds_ids = ds.get_ids().tolist()
        remaining_indexes = [n for n in range(len(ds_ids)) if ds_ids[n] not in processed_ids]

        with torch.no_grad():
            if progress:
                remaining_indexes = tqdm.tqdm(remaining_indexes)
            for n in remaining_indexes:
                img, label, img_path = ds[n]
                if img.dim() == 3:
                    img = img[None, :, :, :]
                elif img.dim() == 2:
                    img = img[None, None, :, :]
                elif img.dim() == 4:
                    pass
                else:
                    raise ValueError(f"Invalid image shape {img.shape}")
                img = img.to(device)
                try:
                    features = []
                    for embeder in embeders:
                        features.append(embeder.forward_sliced(img, 200).detach().cpu()[0, :, 0, 0])
                    features = torch.cat(features, dim=0)
                    features_labels_filenames.append((features, label, img_path))
                except Exception:
                    print(f"Error in embedding {img_path}", file=sys.stderr)
                if n % save_freq == 0:
                    torch.save(features_labels_filenames, resume_path)
            torch.save(features_labels_filenames, resume_path)

    def __init__(self, radii, offset_pattern="8r{}", **kwargs):
        super().__init__()
        if "pooling" in kwargs:
            if kwargs["pooling"] is None or kwargs["pooling"] is False:
                raise ValueError("A MultiRadiusDiffLBP needs an effective pooling strategy.")
        radii_modules = []
        for radius in radii:
            radii_modules.append(DiffLBP(offsets=offset_pattern.format(radius), **kwargs))
        self.diff_lbp = torch.nn.ModuleList(radii_modules)

    def forward(self, x):
        return torch.cat([diff_lbp(x) for diff_lbp in self.diff_lbp], dim=1)

    def forward_sliced(self, x, col_width=200):
        with torch.no_grad():
            return torch.cat([diff_lbp.forward_sliced(x, col_width) for diff_lbp in self.diff_lbp], dim=1)

    def compute_lbp_stack(self, img, return_confidences=False, output_device='cpu')\
            -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        by_radius = [lbp.compute_lbp_image(img, return_confidences=return_confidences) for lbp in self.diff_lbp]
        if return_confidences:
            by_radius = [(lbp_img.to(output_device), conf.to(output_device)) for lbp_img, conf in by_radius]
            lbp_stack = torch.cat([lbp_img for lbp_img, _ in by_radius], dim=1)
            conf_stack = torch.cat([conf for _, conf in by_radius], dim=1)
            return lbp_stack, conf_stack
        else:
            by_radius = [lbp_img.to(output_device) for lbp_img in by_radius]
            lbp_stack = torch.cat(by_radius, dim=1)
            return lbp_stack

    @property
    def radii_count(self) -> int:
        return len(self.diff_lbp)

    @property
    def histograms_size(self):
        return sum(diff_lbp.histogram_size for diff_lbp in self.diff_lbp)

    @property
    def device(self):
        return next(self.parameters()).device


class DropConnect(torch.nn.Module):
    def __init__(self, p=0.1):
        super(DropConnect, self).__init__()
        self.p = p

    def forward(self, x):
        if self.training:
            mask = torch.rand((x.size(0), 1), device=x.device) > self.p
            return x * mask
        return x


class CompressionLayer(torch.nn.Module):
    def __init__(self, input_size, output_size, hellinger_normalise=False,
                 l2_normalise=False, in_dropout=0.0, out_dropconnect=0.0):
        super().__init__()
        self.in_do = torch.nn.Dropout(in_dropout)
        self.fc1 = torch.nn.Linear(input_size, output_size)
        self.out_do = DropConnect(out_dropconnect)
        self.hellinger_normalise = hellinger_normalise
        self.l2_normalise = l2_normalise
        self.eval()

    def forward(self, x):
        x = self.in_do(x)
        x = self.fc1(x)
        x = self.out_do(x)
        if self.hellinger_normalise:
            x = hellinger_normalise(x)
        if self.l2_normalise:
            x = l2_normalise(x)
        return x

    def train_pca(self, x):
        with torch.no_grad():
            sklearn_pca = PCA(n_components=self.fc1.weight.shape[0])
            count = 0
            # replicating sample + noise to avoid singular matrix
            # maybe getting the number of samples from the number of features is a better idea
            while x.shape[0] < self.fc1.weight.shape[1] and count < 5:
                x = torch.cat([x, x + .0000001 * torch.rand_like(x) * .5], dim=0)
            sklearn_pca.fit(x.cpu().detach().numpy())
            self.fc1.weight.data[:, :] = torch.tensor(sklearn_pca.components_).to(self.fc1.weight.device)
            print("PCA training done!", file=sys.stderr)


class DeepCompressionLayer(torch.nn.Module):
    def __init__(self, input_size, output_size, hellinger_normalise=False,
                 l2_normalise=False, in_dropout=0.0, out_dropconnect=0.0):
        super().__init__()
        self.in_do = torch.nn.Dropout(in_dropout)
        self.fc1 = torch.nn.Linear(input_size, output_size)
        self.fc2 = torch.nn.Linear(output_size, output_size)
        self.fc3 = torch.nn.Linear(output_size, output_size)
        self.out_do = DropConnect(out_dropconnect)
        self.hellinger_normalise = hellinger_normalise
        self.l2_normalise = l2_normalise
        self.eval()

    def forward(self, x):
        x = self.in_do(x)
        x = torch.tanh(self.fc1(x))
        x = self.in_do(x)
        x = torch.tanh(self.fc2(x))
        x = self.in_do(x)
        x = self.fc3(x)
        x = self.out_do(x)
        if self.hellinger_normalise:
            x = hellinger_normalise(x)
        if self.l2_normalise:
            x = l2_normalise(x)
        return x

    def train_pca(self, x):
        raise NotImplementedError("Not implemented for DeepCompressionLayer")
        with torch.no_grad():
            sklearn_pca = PCA(n_components=self.fc1.weight.shape[0])
            count = 0
            # replicating sample + noise to avoid singular matrix
            # maybe getting the number of samples from the number of features is a better idea
            while x.shape[0] < self.fc1.weight.shape[1] and count < 5:
                x = torch.cat([x, x + .0000001 * torch.rand_like(x) * .5], dim=0)
            sklearn_pca.fit(x.cpu().detach().numpy())
            self.fc1.weight.data[:, :] = torch.tensor(sklearn_pca.components_).to(self.fc1.weight.device)
            print("PCA training done!", file=sys.stderr)


class HighwayBlock(torch.nn.Module):
    def __init__(self, size, dropout_rate, activation_function=torch.nn.Tanh(), init_bias=-2.0, batch_norm=True):
        super(HighwayBlock, self).__init__()
        self.size = size
        self.activation_function = activation_function
        self.H = torch.nn.Linear(size, size)
        self.T = torch.nn.Linear(size, size)
        self.T.bias.data.fill_(init_bias)
        self.dropout = torch.nn.Dropout(dropout_rate)
        self.batch_norm = torch.nn.BatchNorm1d(size) if batch_norm else None

    def forward(self, x):
        H_out = self.activation_function(self.H(x))
        H_out = self.dropout(H_out)
        T_out = torch.sigmoid(self.T(x))
        if self.batch_norm:
            H_out = self.batch_norm(H_out)
        C_out = 1 - T_out
        return H_out * T_out + x * C_out


class DeepResCompressionLayer(torch.nn.Module):
    def __init__(self, input_size, output_size, depth=5, hellinger_normalise=False,
                 l2_normalise=False, in_dropout=0.0, out_dropconnect=0.0, init_highway_bias=-2.0):
        super().__init__()
        self.in_do = torch.nn.Dropout(in_dropout)
        self.fc1 = torch.nn.Linear(input_size, output_size)
        highway_blocks = [HighwayBlock(output_size, in_dropout, init_bias=init_highway_bias) for _ in range(depth)]
        self.highway_blocks = torch.nn.Sequential(*highway_blocks)

        self.out_do = DropConnect(out_dropconnect)
        self.hellinger_normalise = hellinger_normalise
        self.l2_normalise = l2_normalise
        self.eval()

    def forward(self, x):
        x = self.in_do(torch.tanh(self.fc1(x)))
        x = self.highway_blocks(x)
        x = self.out_do(x)
        if self.hellinger_normalise:
            x = hellinger_normalise(x)
        if self.l2_normalise:
            x = l2_normalise(x)
        return x

    def train_pca(self, x):
        raise NotImplementedError("Not implemented for DeepCompressionLayer")


class DiffSRSLBP(torch.nn.Module):
    def __init__(self, radii=list(range(1, 12)), num_components=200, **kwargs) -> None:
        super().__init__()
        self.multi_radius_diff_lbp = MultiRadiusDiffLBP(radii, **kwargs)
        self.compression_layer = CompressionLayer(self.multi_radius_diff_lbp.histograms_size, num_components)
        self.eval()

    def forward(self, x):
        x = self.multi_radius_diff_lbp(x)
        x = self.compression_layer(x)
        return x

    def train_pca(self, data):
        self.compression_layer.train_pca(data)



import pytest
import ptlbp
import torch
import numpy as np
from ptlbp.testing import generate_noise_image, generate_image


debug_test_multiradii = True


@pytest.mark.parametrize("img, radii_count, num_components, out_std", [
    [generate_image(100), 6, 200, .0019],
    [generate_image(100), 3, 200, .0028],
    [generate_image(100), 2, 200, .0033],
    [generate_image(100), 1, 200, .0052],
    [generate_noise_image(100), 1, 200, .0026],
    [generate_noise_image(100), 1, 0, 0.0065]  # No PCA and post-PCA processing
    ])
def test_multiradii(img, radii_count, num_components, out_std):
    torch.manual_seed(1337)
    lbp = ptlbp.DiffSRSLBP(num_components=num_components, radii=range(1, radii_count + 1))
    output = lbp(img)
    if num_components > 0:
        assert output.size(0) == 1 and output.size(1) == num_components and output.size(2) == output.size(3) == 1
    else:
        assert output.size(0) == 1 and output.size(1) == 254 * radii_count and output.size(2) == output.size(3) == 1
    assert np.allclose(output.detach().numpy().std(), out_std, atol=.00015)


#  Test the pca_normalise function
#  Essentially test that after learning the pca components, embeedings from images with similar texture are closer
#  then ones with different texture on 10 reclications and with a margin of other_dist - similar_dist
@pytest.mark.parametrize("mode_balance, far_mode_balance, margin", [
    [.1, .9, .1],
    [.2, .8, .1],
    [.3, .7, .05],
    [.4, .6, .03],
    [.1, .4, .1],  # There is no margin here because the far_mode_balance is too close to the mode_balance
])
def test_train_pca(mode_balance, far_mode_balance, margin):

    def dist(x, y):
        return (((x.view(-1)-y.view(-1))**2).sum()**.5).detach().item()

    lbp = ptlbp.DiffSRSLBP(num_components=200, radii=range(1, 6))
    dataloader = torch.utils.data.DataLoader(
        [generate_noise_image(100, mode_balance=.2+(s/10.) * .6, seed=s)[0] for s in range(10)], batch_size=1)
    lbp.train_pca(dataloader)
    for seed_base in [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000]:
        ancor = lbp(generate_noise_image(100, mode_balance=mode_balance, seed=seed_base))
        similar = lbp(generate_noise_image(100, mode_balance=mode_balance, seed=seed_base + 1))
        other = lbp(generate_noise_image(100, mode_balance=far_mode_balance, seed=seed_base + 2))
        assert dist(ancor, similar) + margin < dist(ancor, other)


def test_gpu():
    lbp = ptlbp.DiffSRSLBP(num_components=200, radii=range(1, 6))
    img = generate_noise_image(100, mode_balance=.5, seed=1337)
    output = lbp(img)
    if torch.cuda.is_available():
        lbp.to('cuda')
        cuda_output = lbp(img.to('cuda'))
        assert torch.allclose(output, cuda_output.cpu(), atol=1e-6)
    else:
        with pytest.raises(RuntimeError):
            lbp.to('cuda')
            lbp(generate_noise_image(100, mode_balance=.5, seed=1337).to('cuda'))



import torch
import numpy as np
#  Usefull for testing:


def generate_rect_image(sz) -> torch.Tensor:
    image = torch.zeros([1, 1, sz, sz], dtype=torch.float32)
    image[:, :, sz//4:3*sz//4, sz//4:3*sz//4] = 1
    return image


def generate_cross_image(sz) -> torch.Tensor:
    image = torch.zeros([1, 1, sz, sz], dtype=torch.float32)
    image[:, :, ((3*sz)//8): ((5*sz)//8), ((1*sz)//4):((3*sz)//4)] = 1.
    image[:, :, ((1*sz)//4):((3*sz)//4), ((3*sz)//8): ((5*sz)//8)] = 1.
    return image


def generate_cross_and_small():
    img = np.zeros([100, 100])
    img[20:80, 40:60] = 1  # 5x20x20 = 2000 pixels
    img[40:60, 20:80] = 1
    img[5, 5] = 1
    img[5, 93:96] = 1
    img[93:96, 5] = 1
    return torch.tensor(img[None, None, :, :], dtype=torch.float32)


def generate_square_and_small():
    img = np.zeros([100, 100])
    img[20:80, 40:60] = 1  # 40x50 = 2000 pixels
    img[5, 5] = 1
    img[5, 93:96] = 1
    img[93:96, 5] = 1
    return torch.tensor(img[None, None, :, :], dtype=torch.float32)


def generate_noise_image(sz, mode_balance=.5, small_noise=.1, seed=1337):
    if isinstance(sz, int):
        sz = (sz, sz)
    assert isinstance(sz, tuple)
    torch.manual_seed(seed)
    img = (torch.rand(*sz) > mode_balance).float()
    img[sz[1]//10: 2 * (sz[1]//10), 2 * sz[0]//10: 8 * (sz[0]//10)] = 1
    img[7 * (sz[1]//10): 8 * (sz[1]//10), 2 * sz[0]//10: 8 * (sz[0]//10)] = 0
    img = img * (1-small_noise) + small_noise * torch.rand_like(img)
    return img[None, None, :, :]


def generate_bimodal_image(sz=100, mode1_mean_std=(.4, .2),
                           mode2_mean_std=(.8, .1), mode1_prob=.8, seed=1337) -> torch.Tensor:
    if isinstance(sz, int):
        sz = (sz, sz)
    assert isinstance(sz, tuple)
    torch.manual_seed(seed)
    mode1_pixels = torch.normal(torch.ones(sz) * mode1_mean_std[0], torch.ones(sz) * mode2_mean_std[1])
    mode2_pixels = torch.normal(torch.ones(sz) * mode2_mean_std[0], torch.ones(sz) * mode2_mean_std[1])
    choice = (torch.rand(sz) < mode1_prob).float()
    img = mode1_pixels * choice + mode2_pixels * (1 - choice)
    return img[None, None, :, :]


def generate_image(sz=100):
    if isinstance(sz, int):
        sz = (sz, sz)
    assert isinstance(sz, tuple)
    w, h = sz
    img = torch.zeros(sz)
    img[h//10, w//10] = 1
    img[h//10 - 2: h//10 + 2, 9 * (w//10)] = 1
    img[9 * (h//10), w//10-2: w//10+2] = 1

    img[2*(h//10): 8*(h//10), 4*(w//10): 6*(w//10)] = 1
    img[4*(h//10): 6*(h//10), 2*(w//10): 8*(w//10)] = 1

    img[2*(h//10): 5*(h//10), 2*(w//10): 3*(w//10)] = 1
    img[2*(h//10): 3*(h//10), 2*(w//10): 5*(w//10)] = 1
    return img[None, None, :, :]



from typing import Any, Dict, Tuple, Union, List
from .srs import MultiRadiusDiffLBP
from .util import pil_image_as_batch
from PIL import Image
import numpy as np
import torch
from collections import defaultdict

TPattern = Union[np.uint64, str, Tuple[int, ...], List[int], np.ndarray, torch.tensor]
TImage = Union[Image.Image, np.ndarray, torch.Tensor, None]


class DenseLBPIndex():
    def __add__(self, other: 'DenseLBPIndex') -> 'DenseLBPIndex':
        assert self.lbp_extractor == other.lbp_extractor and\
            self._do_cache_images == other._do_cache_images and\
            self._do_cache_occurences == other._do_cache_occurences\
            and self._do_cache_stacks == other._do_cache_stacks
        res = DenseLBPIndex(lbp_extractor=self.lbp_extractor,
                            cache_images=self._do_cache_images,
                            cache_occurences=self._do_cache_occurences,
                            cache_stacks=self._do_cache_stacks)
        for lbp_img in self._dense_lbp_images:
            res._dense_lbp_images.append(lbp_img)
            res._ids.append(lbp_img.id)
            res._ids_to_index[lbp_img.id] = len(res._dense_lbp_images) - 1
        for lbp_img in other._dense_lbp_images:
            if lbp_img.id not in self._ids_to_index:
                res._dense_lbp_images.append(lbp_img)
                res._ids.append(lbp_img.id)
                res._ids_to_index[lbp_img.id] = len(res._dense_lbp_images) - 1
            else:
                print(f"__add__ Image with id {lbp_img.id} already exists")
        return res

    def pattern_to_string(self, pattern: TPattern) -> str:
        if isinstance(pattern, str):
            return pattern
        elif isinstance(pattern, (tuple, list, np.ndarray, torch.Tensor)):
            return f"({','.join([str(int(x)) for x in pattern])})"
        else:
            raise ValueError(f'Invalid pattern type {type(pattern)}')

    def pattern_to_tuple(self, pattern: TPattern) -> torch.Tensor:
        if isinstance(pattern, str):
            return tuple([int(p) for p in pattern.strip('()').split(',')])
        elif isinstance(pattern, tuple):
            return pattern
        elif isinstance(pattern, list):
            return tuple(pattern)
        elif isinstance(pattern, np.ndarray):
            return tuple(pattern.reshape(-1))
        elif isinstance(pattern, torch.Tensor):
            return tuple(pattern.cpu().numpy().reshape(-1))
        else:
            raise ValueError(f'Invalid pattern type {type(pattern)}')

    def get_pattern_frequecies(self, return_str: bool = False) -> Union[Dict[str, int], Dict[Tuple[int, ...], int]]:
        occ, _ = self.get_occurence_count()
        freq_pattern = sorted([(v, self.pattern_to_string(k)) for k, v in occ.items()])
        if return_str:
            res = {self.pattern_to_string(p): f for f, p in freq_pattern}
        else:
            res = {self.pattern_to_tuple(p): f for f, p in freq_pattern}
        return res

    def __init__(self, lbp_extractor: MultiRadiusDiffLBP, cache_images: bool = True,
                 cache_stacks: bool = False, cache_occurences: bool = True) -> None:
        if not cache_images and not cache_occurences:
            raise ValueError('Cannot cache neither occurences nor images')
        self.lbp_extractor = lbp_extractor
        self._dense_lbp_images = []
        self._ids = []
        self._ids_to_index: Dict[str, int] = {}
        self._cached_global_occurences = None
        self._cached_per_image_occurences = None
        self._do_cache_stacks = cache_stacks
        self._do_cache_images = cache_images
        self._do_cache_occurences = cache_occurences

    @property
    def radii_count(self) -> int:
        return self.lbp_extractor.radii_count

    def add_image(self, img_id: str, img: TImage = None) -> 'DenseLBPIndex':
        if img_id in self._ids_to_index:
            raise ValueError('Image with id already exists')
        if img is None:
            img = Image.open(img_id)
        lbp_image = DenseImageLBP(id=img_id, img=img, lbp_extractor=self.lbp_extractor,
                                  cache_images=self._do_cache_images, cache_occurences=self._do_cache_occurences,
                                  cache_stack=self._do_cache_stacks)
        self._dense_lbp_images.append(lbp_image)
        self._ids.append(img_id)
        self._ids_to_index[img_id] = len(self._dense_lbp_images) - 1
        self._cached_global_occurences = None
        self._cached_per_image_occurences = None
        return self

    def add_images(self, image_paths) -> 'DenseLBPIndex':
        for image_path in image_paths:
            self.add_image(image_path)
        return self

    def __getitem__(self, pos_or_id: Union[int, str]) -> 'DenseImageLBP':
        if isinstance(pos_or_id, int):
            return self._dense_lbp_images[pos_or_id]
        elif isinstance(pos_or_id, str):
            return self._dense_lbp_images[self._ids_to_index[pos_or_id]]
        else:
            raise ValueError('Invalid index type')

    def __len__(self):
        return len(self._dense_lbp_images)

    def get_occurence_count(self) -> \
            Tuple[Dict[Tuple[int, ...], int], Dict[str, Dict[Tuple[int, ...], int]]]:
        if self._cached_global_occurences is not None and self._cached_per_image_occurences is not None:
            return self._cached_global_occurences, self._cached_per_image_occurences
        global_occurences = defaultdict(lambda: 0)
        per_image_occurences = {}
        for indexed_image in self._dense_lbp_images:
            occurences = indexed_image.occurences
            per_image_occurences[indexed_image.id] = occurences
            for pattern, count in occurences.items():
                global_occurences[pattern] += count
        self._cached_global_occurences = dict(global_occurences)
        self._cached_per_image_occurences = per_image_occurences
        return self._cached_global_occurences, self._cached_per_image_occurences


class FastLBPIndex(DenseLBPIndex):
    def __add__(self, other: 'FastLBPIndex') -> 'FastLBPIndex':
        assert self.lbp_extractor == other.lbp_extractor and\
            self._do_cache_images == other._do_cache_images and\
            self._do_cache_occurences == other._do_cache_occurences\
            and self._do_cache_stacks == other._do_cache_stacks
        res = FastLBPIndex(lbp_extractor=self.lbp_extractor,
                           cache_images=self._do_cache_images,
                           cache_occurences=self._do_cache_occurences,
                           cache_stacks=self._do_cache_stacks)
        for lbp_img in self._dense_lbp_images:
            res._dense_lbp_images.append(lbp_img)
            res._ids.append(lbp_img.id)
            res._ids_to_index[lbp_img.id] = len(res._dense_lbp_images) - 1
        for lbp_img in other._dense_lbp_images:
            if lbp_img.id not in self._ids_to_index:
                res._dense_lbp_images.append(lbp_img)
                res._ids.append(lbp_img.id)
                res._ids_to_index[lbp_img.id] = len(res._dense_lbp_images) - 1
            else:
                print(f"__add__ Image with id {lbp_img.id} already exists")
        return res

    def pattern_to_string(self, pattern: TPattern) -> str:
        if isinstance(pattern, str):
            return pattern
        elif isinstance(pattern, np.uint64):
            return f"{repr(self.decode_pattern(pattern))}"
        elif isinstance(pattern, (tuple, list, np.ndarray, torch.Tensor)):
            return f"({','.join([str(int(x)) for x in pattern])})"
        else:
            raise ValueError(f'Invalid pattern type {type(pattern)}')

    def pattern_to_tuple(self, pattern: TPattern) -> torch.Tensor:
        if isinstance(pattern, str):
            return tuple([int(p) for p in pattern.strip('()').split(',')])
        elif isinstance(pattern, np.uint64):
            return self.decode_pattern(pattern)
        elif isinstance(pattern, tuple):
            return pattern
        elif isinstance(pattern, list):
            return tuple(pattern)
        elif isinstance(pattern, np.ndarray):
            return tuple(pattern.reshape(-1))
        elif isinstance(pattern, torch.Tensor):
            return tuple(pattern.cpu().numpy().reshape(-1))
        else:
            raise ValueError(f'Invalid pattern type {type(pattern)}')

    def encode_pattern(self, pattern: TPattern) -> int:
        pattern = np.array(pattern).reshape(-1)
        if len(pattern) != self.radii_count:
            raise ValueError('Pattern must be convertible to 1D array')
        return (self._pattern_exponents * pattern).sum()

    def decode_pattern(self, pattern: np.uint64) -> Tuple[int, ...]:
        return tuple((pattern // self._pattern_exponents) % 256)

    def __init__(self, lbp_extractor: MultiRadiusDiffLBP, cache_images: bool = True,
                 cache_stacks: bool = False, cache_occurences: bool = True) -> None:
        super().__init__(lbp_extractor=lbp_extractor, cache_images=cache_images, cache_stacks=cache_stacks,
                         cache_occurences=cache_occurences)
        if lbp_extractor.radii_count > 8:
            raise ValueError('FastLBPIndex only supports radii count <= 8')
        self._pattern_exponents = np.array([256 ** i for i in range(lbp_extractor.radii_count)], dtype=np.uint64)

    @property
    def radii_count(self) -> int:
        return self.lbp_extractor.radii_count

    def add_image(self, img_id: str, img: TImage = None) -> 'DenseLBPIndex':
        if img_id in self._ids_to_index:
            raise ValueError('Image with id already exists')
        if img is None:
            img = Image.open(img_id)
        lbp_image = FastImageLBP(id=img_id, img=img, lbp_extractor=self.lbp_extractor,
                                 cache_images=self._do_cache_images, cache_occurences=self._do_cache_occurences,
                                 cache_stack=self._do_cache_stacks)
        self._dense_lbp_images.append(lbp_image)
        self._ids.append(img_id)
        self._ids_to_index[img_id] = len(self._dense_lbp_images) - 1
        self._cached_global_occurences = None
        self._cached_per_image_occurences = None
        return self

    def add_images(self, image_paths) -> 'DenseLBPIndex':
        for image_path in image_paths:
            self.add_image(image_path)
        return self

    def __getitem__(self, pos_or_id: Union[int, str]) -> 'DenseImageLBP':
        if isinstance(pos_or_id, int):
            return self._dense_lbp_images[pos_or_id]
        elif isinstance(pos_or_id, str):
            return self._dense_lbp_images[self._ids_to_index[pos_or_id]]
        else:
            raise ValueError('Invalid index type')

    def get_occurence_count(self) -> \
            Tuple[Dict[int, int], Dict[str, Dict[int, int]]]:
        if self._cached_global_occurences is not None and self._cached_per_image_occurences is not None:
            return self._cached_global_occurences, self._cached_per_image_occurences
        global_occurences = defaultdict(lambda: 0)
        per_image_occurences = {}
        for indexed_image in self._dense_lbp_images:
            occurences = indexed_image.occurences
            per_image_occurences[indexed_image.id] = occurences
            for pattern, count in occurences.items():
                global_occurences[pattern] += count
        self._cached_global_occurences = dict(global_occurences)
        self._cached_per_image_occurences = per_image_occurences
        return self._cached_global_occurences, self._cached_per_image_occurences


class DenseImageLBP():
    @staticmethod
    def resolve_img_type(img: TImage) -> None:
        if isinstance(img, torch.Tensor):
            pt_img = img
            if pt_img.dim() == 3:
                pt_img = pt_img.unsqueeze(0)
            elif pt_img.dim() == 2:
                pt_img = pt_img.unsqueeze(0).unsqueeze(0)
            elif pt_img.dim() != 4:
                raise ValueError('Invalid tensor shape')
            pil_img = Image.fromarray((img[0, 0, :, :].cpu().numpy() * 255.).astype(np.uint8))
        elif isinstance(img, np.ndarray):
            pt_img = torch.tensor(img)
            if pt_img.dim() == 3:
                pt_img = pt_img.unsqueeze(0)
            elif pt_img.dim() == 2:
                pt_img = pt_img.unsqueeze(0).unsqueeze(0)
            elif pt_img.dim() != 4:
                raise ValueError('Invalid tensor shape')
            pil_img = Image.fromarray((img[0, 0, :, :].cpu().numpy() * 255.).astype(np.uint8))
        elif isinstance(img, Image.Image):
            pil_img = img
            pt_img = pil_image_as_batch(img)
        elif isinstance(img, str):
            pil_img = Image.open(img)
            pt_img = pil_image_as_batch(pil_img)
        else:
            raise ValueError(f"Invalid image type {type(img)}")
        return pil_img, pt_img

    def __init__(self, id: str, lbp_extractor: MultiRadiusDiffLBP, img: TImage = None, cache_stack: bool = True,
                 cache_occurences: bool = True, cache_images: bool = True) -> None:
        if not cache_images and not cache_occurences:
            raise ValueError('Cannot cache neither occurences nor images')
        self._lbp_extractor = lbp_extractor
        self._id = id
        self._img = None
        self._cached_stack = None
        self._cached_occurences = None
        self._do_cache_occurences = cache_occurences
        self._do_cache_stack = cache_stack
        if img is None:
            img = id
            pil_img, pt_img = DenseImageLBP.resolve_img_type(img=img)
            self._img = pil_img
        else:
            self._img = img

    @property
    def radii_count(self) -> int:
        return self._lbp_extractor.radii_count

    @property
    def computation_device(self) -> torch.device:
        return self._lbp_extractor.device

    @property
    def id(self) -> str:
        return self._id

    @property
    def stack(self) -> Union[torch.Tensor, None]:
        if self._cached_stack is None:
            _, pt_img = DenseImageLBP.resolve_img_type(img=self._img)
            stack = self._compute_stack(pt_img.to(self.computation_device))
            if self._do_cache_stack:
                self._cached_stack = stack
            return stack
        else:
            return self._cached_stack

    @property
    def occurences(self) -> Dict[Tuple[int, ...], int]:
        stack = self.stack  # caching in case it is not cached
        if self._cached_occurences is None:
            occurences = defaultdict(lambda: 0)
            for x in range(stack.shape[1]):
                for y in range(stack.shape[2]):
                    pattern = tuple(stack[:, x, y])
                    occurences[pattern] += 1
            occurences = dict(occurences)
            if self._do_cache_occurences:
                self._cached_occurences = occurences
            return occurences
        else:
            return self._cached_occurences

    @property
    def img(self) -> Union[Image.Image, None]:
        return self._img

    def _compute_stack(self, pt_img: torch.Tensor) -> np.ndarray:
        res = self._lbp_extractor.compute_lbp_stack(pt_img.to(self.computation_device), return_confidences=False)
        res = res.cpu().numpy().astype(int)
        if len(res.shape) == 4 and res.shape[0] == 1:
            res = res.squeeze(0)
        return res

    def get_similarity_count(self, pattern: TPattern) -> int:
        pattern = np.array(pattern).reshape(-1)[:, None, None]
        if len(pattern.shape) != 3 or pattern.shape[0] != self.radii_count:
            raise ValueError('Pattern must be convertible to 3D array')
        agree = (self.stack == pattern).astype(int).sum(axis=0)
        return agree

    def locate_pattern(self, pattern: TPattern, tolerance=0) -> Tuple[np.ndarray, np.ndarray]:
        correct = self.get_similarity_count(pattern=pattern) >= (self.radii_count - tolerance)
        y, x = np.nonzero(correct)
        return x, y


class FastImageLBP(DenseImageLBP):
    def encode_pattern(self, pattern: TPattern) -> int:
        pattern = np.array(pattern).reshape(-1)
        if len(pattern) != self.radii_count:
            raise ValueError('Pattern must be convertible to 1D array')
        return (self._pattern_exponents * pattern).sum()

    def decode_pattern(self, pattern: int) -> Tuple[int, ...]:
        return tuple((pattern // self._pattern_exponents) % 256)

    def __init__(self, id: str, lbp_extractor: MultiRadiusDiffLBP, img: TImage = None, cache_stack: bool = True,
                 cache_occurences: bool = True, cache_images: bool = True) -> None:
        super().__init__(id=id, lbp_extractor=lbp_extractor, img=img, cache_stack=cache_stack,
                         cache_occurences=cache_occurences, cache_images=cache_images)
        assert lbp_extractor.radii_count <= 8
        self._pattern_exponents = np.array([256 ** i for i in range(lbp_extractor.radii_count)], dtype=np.uint64)

    @property
    def radii_count(self) -> int:
        return self._lbp_extractor.radii_count

    @property
    def computation_device(self) -> torch.device:
        return self._lbp_extractor.device

    @property
    def id(self) -> str:
        return self._id

    @property
    def stack(self) -> Union[torch.Tensor, None]:
        if self._cached_stack is None:
            _, pt_img = DenseImageLBP.resolve_img_type(img=self._img)
            stack = self._compute_stack(pt_img.to(self.computation_device))
            if self._do_cache_stack:
                #print("Caching stack", type(self._cached_stack))
                self._cached_stack = stack
            else:
                pass
                #print("Not Caching stack", type(self._cached_stack))
            return stack
        else:
            return self._cached_stack

    @property
    def occurences(self) -> Dict[int, int]:
        if self._cached_occurences is None:
            stack = self.stack
            patterns, freqs = np.unique(stack, return_counts=True)
            occurences = dict(zip(patterns, freqs))
            if self._do_cache_occurences:
                self._cached_occurences = occurences
            return occurences
        else:
            return self._cached_occurences

    @property
    def img(self) -> Union[Image.Image, None]:
        return self._img

    def _compute_stack(self, pt_img: torch.Tensor) -> np.ndarray:
        stack = self._lbp_extractor.compute_lbp_stack(pt_img.to(self.computation_device), return_confidences=False)
        if len(stack.shape) == 4 and stack.shape[0] == 1:
            stack = stack.squeeze(0)
        stack = (stack.cpu().numpy().astype(np.uint64) * self._pattern_exponents[:, None, None]).sum(axis=0)
        return stack

    def get_similarity_count(self, pattern: TPattern) -> int:
        pattern = self.encode_pattern(pattern)
        if len(pattern.shape) != 3 or pattern.shape[0] != self.radii_count:
            raise ValueError('Pattern must be convertible to 3D array')
        agree = (self.stack == pattern)
        return agree

    def locate_pattern(self, pattern: TPattern, tolerance=0) -> Tuple[np.ndarray, np.ndarray]:
        if tolerance > 0:
            raise NotImplementedError('Tolerance not implemented for FastImageLBP')
        correct = self.get_similarity_count(pattern=pattern) >= (self.radii_count - tolerance)
        y, x = np.nonzero(correct)
        return x, y



import pytest
import torch
import numpy as np
from ptlbp import _DiffBinning, DiffHistogram, DiffOtsu, hard_otsu_threshold, diff_otsu_threshold
import ptlbp
from ptlbp.testing import generate_bimodal_image
from ptlbp.util import np_image_as_batch, plot_img
from skimage import data
from skimage.filters import threshold_otsu as sk_otsu

np.set_printoptions(linewidth=400)


debug_binning = False
debug_hard_otsu = False
if any([debug_binning, debug_hard_otsu]):
    from matplotlib import pyplot as plt


def test_diffbinning():
    img = np_image_as_batch(data.coins())
    sz = list(img.size())
    bins = torch.linspace(-0.1, 1.1, 100)
    binned = _DiffBinning(bin_centers=bins)(img)
    assert binned.size() == torch.Size(sz + [bins.size(0)])
    binned_sum = binned.sum(dim=-1)
    if debug_binning:
        print(torch.aminmax(img))
        plt.imshow(img[0, 0, :, :].numpy())
        plt.title(f"Input image {torch.aminmax(img)}")
        plt.colorbar()
        plt.show(block=False)
        plt.figure()
        plt.imshow(binned_sum[0, 0, :, :].numpy())
        plt.title(f"Binned sum {torch.aminmax(binned_sum)}")
        plt.colorbar()
        plt.show()
    assert torch.allclose(binned.sum(dim=-1), torch.ones(sz), atol=1e-5)
    # Reconstruction only works well when no clamping is needed
    reconstructed = (binned * bins.view(*([1] * img.dim() + [-1]))).sum(dim=-1)
    assert torch.allclose(reconstructed, img, atol=1e-5)


@pytest.mark.parametrize(["img"], [[data.coins()], [data.camera()], [data.page()],
                                   [np.zeros([100, 100])], [np.ones([100, 100])],
                                   [(np.random.rand(100, 100) * 255).astype(np.uint8)]])
def test_hard_otsu(img):
    img = np_image_as_batch(img)
    hist = torch.histc(img, bins=100, min=0, max=1)
    otsu_thr = hard_otsu_threshold(hist).item() / 100
    sk_thr = sk_otsu(img[0, 0, :, :].numpy())
    if debug_hard_otsu:
        print(f"Mean Hard: {(img > otsu_thr).float().mean().item()} Mean SK{(img > otsu_thr).float().mean().item()}")
        plot_img(img, title="Input image", block=False)
        plot_img(img > otsu_thr, title="My Otsu", block=False)
        plot_img(img > sk_thr, title="SK Otsu", block=True)
    assert (img > otsu_thr).float().mean().item() == (img > otsu_thr).float().mean().item()



from .sampling import PointSampler, GaussianPointSampler, TOffsets
from .comparisson import Comparisson, ComparissonQuantile, ComparissonOtsu
from .encoding import BinaryToOnehotChannelEmbeddings
from .diff_lbp import DiffLBP
from .srs import DiffSRSLBP, CompressionLayer, DeepCompressionLayer, DeepResCompressionLayer
from .srs import DropConnect, MultiRadiusDiffLBP
from .srs import hellinger_normalise, l2_normalise
from .losses import CDistLoss, CDistContrastiveLoss, CDistTripletLoss, CDistRatioLoss
from .histograms import DiffHOG, DiffHistogram, DiffOtsu, hard_otsu_threshold, diff_otsu_threshold, _DiffBinning
from . import unpooled_lbp
from . import util
from . import testing
from . import datasets


__all__ = ["DiffLBP", "PointSampler", "BinaryToOnehotChannelEmbeddings", "Comparisson", "ComparissonOtsu",
           "ComparissonQuantile", "util", "testing", "TOffsets", "DiffSRSLBP", "CompressionLayer",
           "DeepCompressionLayer", "DeepResCompressionLayer", "DropConnect", "MultiRadiusDiffLBP",
           "hellinger_normalise", "l2_normalise", "CDistLoss", "CDistContrastiveLoss", "CDistTripletLoss",
           "CDistRatioLoss", "DiffHOG", "DiffHistogram", "DiffOtsu", "hard_otsu_threshold",
           "diff_otsu_threshold", "_DiffBinning", "datasets", "unpooled_lbp", "GaussianPointSampler"]


__version__ = "0.1.0"

# redistering comlormap 'd256' to matplotlib
try:
    util.create_dicrete_colormap(256)
except ValueError:
    # colormap already registered lets fail gracefully in __init__
    pass


# TODO: add retrieval searcher for the classes
